# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

from typing import List, Dict, Union, Any

from pathlib import Path
import numpy as np

import odatse
from ._domain import DomainBase

class Region(DomainBase):
    """
    A class to represent a region in the domain.

    Attributes
    ----------
    min_list : np.array
        Minimum values for each dimension.
    max_list : np.array
        Maximum values for each dimension.
    unit_list : np.array
        Unit values for each dimension.
    initial_list : np.array
        Initial values for each dimension.
    """

    min_list: np.array
    max_list: np.array
    unit_list: np.array
    initial_list: np.array

    def __init__(self, info: odatse.Info = None, *, param: Dict[str, Any] = None):
        """
        Initialize the Region object.

        Parameters
        ----------
        info : odatse.Info, optional
            Information object containing algorithm parameters.
        param : dict, optional
            Dictionary containing algorithm parameters.
        """
        super().__init__(info)

        if info:
            if "param" in info.algorithm:
                self._setup(info.algorithm["param"])
            else:
                raise ValueError("ERROR: algorithm.param not defined")
        elif param:
            self._setup(param)
        else:
            pass

    def _setup(self, info_param):
        """
        Setup the region with the given parameters.

        Parameters
        ----------
        info_param : dict
            Dictionary containing the parameters for the region.
        """
        if "min_list" not in info_param:
            raise ValueError("ERROR: algorithm.param.min_list is not defined in the input")
        min_list = np.array(info_param["min_list"])

        if "max_list" not in info_param:
            raise ValueError("ERROR: algorithm.param.max_list is not defined in the input")
        max_list = np.array(info_param["max_list"])

        if len(min_list) != len(max_list):
            raise ValueError("ERROR: lengths of min_list and max_list do not match")

        self.dimension = len(min_list)

        unit_list = np.array(info_param.get("unit_list", [1.0] * self.dimension))

        self.min_list = min_list
        self.max_list = max_list
        self.unit_list = unit_list

        initial_list = np.array(info_param.get("initial_list", []))
        if initial_list.ndim == 1:
            initial_list = initial_list.reshape(1, -1)

        if initial_list.size > 0:
            if initial_list.shape[1] != self.dimension:
                raise ValueError("ERROR: dimension of initial_list is incorrect")
            self.num_walkers = initial_list.shape[0]
        else:
            self.num_walkers = 0

        self.initial_list = initial_list

    def initialize(self, rng=np.random, limitation=odatse.util.limitation.Unlimited(), num_walkers: int = 1):
        """
        Initialize the region with random values or predefined initial values.

        Parameters
        ----------
        rng : numpy.random, optional
            Random number generator.
        limitation : odatse.util.limitation, optional
            Limitation object to judge the validity of the values.
        num_walkers : int, optional
            Number of walkers to initialize.
        """
        if num_walkers > self.num_walkers:
            self.num_walkers = num_walkers

        if self.initial_list.size > 0 and self.initial_list.shape[0] >= num_walkers:
            pass
        else:
            self._init_random(rng=rng, limitation=limitation)

    def _init_random(self, rng=np.random, limitation=odatse.util.limitation.Unlimited(), max_count=100):
        """
        Initialize the region with random values within the specified limits.

        Parameters
        ----------
        rng : numpy.random, optional
            Random number generator.
        limitation : odatse.util.limitation, optional
            Limitation object to judge the validity of the values.
        max_count : int, optional
            Maximum number of trials to generate valid values.
        """
        initial_list = np.zeros((self.num_walkers, self.dimension), dtype=float)
        is_ok = np.full(self.num_walkers, False)

        if self.initial_list.size > 0:
            nitem = min(self.num_walkers, self.initial_list.shape[0])
            initial_list[0:nitem] = self.initial_list[0:nitem]
            is_ok[0:nitem] = True

        count = 0
        while (not np.all(is_ok)):
            count += 1
            initial_list[~is_ok] = self.min_list + (self.max_list - self.min_list) * rng.rand(np.count_nonzero(~is_ok), self.dimension)
            is_ok = np.array([limitation.judge(v) for v in initial_list])
            if count >= max_count:
                raise RuntimeError("ERROR: init_random: trial count exceeds {}".format(max_count))
        self.initial_list = initial_list

if __name__ == "__main__":
    reg = Region(param={
        "min_list": [0.0, 0.0, 0.0],
        "max_list": [1.0, 1.0, 1.0],
        "initial_list": [[0.1, 0.2, 0.3],
                         [0.2, 0.3, 0.1],
                         [0.3, 0.1, 0.2],
                         [0.2, 0.1, 0.3],
                         [0.1, 0.3, 0.2],
                         ],
    })

    #lim = odatse.util.limitation.Unlimited()
    lim = odatse.util.limitation.Inequality(a=np.array([[1,0,0],[-1,-1,-1]]),b=np.array([0,1]),is_limitary=True)
    
    reg.initialize(np.random, lim, 8)
    
    print(reg.min_list, reg.max_list, reg.unit_list, reg.initial_list, reg.num_walkers)



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import typing
from typing import TextIO, Union, List, Tuple, Optional, TYPE_CHECKING
import copy
import time
from pathlib import Path

import numpy as np

import odatse
import odatse.domain
from odatse import mpi
from odatse.algorithm.state import ContinuousStateSpace, DiscreteStateSpace
from odatse.util.data_writer import DataWriter

if TYPE_CHECKING:
    from mpi4py import MPI


class AlgorithmBase(odatse.algorithm.AlgorithmBase):
    """
    Base class for Monte Carlo algorithms implementing common functionality.
    
    This class provides the foundation for various Monte Carlo methods, handling both
    continuous and discrete parameter spaces. It supports parallel execution with
    multiple walkers and temperature-based sampling methods.

    Implementation Details
    ----------------------
    The class handles two types of parameter spaces:
      1. Continuous: Uses real-valued parameters within specified bounds
      2. Discrete: Uses node-based parameters with defined neighbor relationships

    For continuous problems:
      - Parameters are bounded by xmin and xmax
      - Steps are controlled by xstep for each dimension
    
    For discrete problems:
      - Parameters are represented as nodes in a graph
      - Transitions are only allowed between neighboring nodes
      - Neighbor relationships must form a connected, bidirectional graph

    The sampling process:
      1. Initializes walkers in valid positions
      2. Proposes moves based on the parameter space type
      3. Evaluates the objective function ("Energy")
      4. Accepts/rejects moves based on the Monte Carlo criterion
      5. Tracks the best solution found

    Key Methods
    -----------
      _initialize() :
         Sets up initial walker positions and counters
      propose() :
         Generates candidate moves for walkers
      local_update() :
         Performs one Monte Carlo step
      _evaluate() :
         Computes objective function values
    """

    nwalkers: int

    iscontinuous: bool

    # # continuous problem
    # x: np.ndarray
    # xmin: np.ndarray
    # xmax: np.ndarray
    # xstep: np.ndarray

    # # discrete problem
    # inode: np.ndarray
    # nnodes: int
    # node_coordinates: np.ndarray
    # neighbor_list: List[List[int]]
    # ncandidates: np.ndarray  # len(neighbor_list[i])-1

    # state: Union[ContinuousState, DiscreteState]

    numsteps: int

    fx: np.ndarray
    istep: int
    best_x: np.ndarray
    best_fx: float
    best_istep: int
    best_iwalker: int
    betas: np.ndarray
    input_as_beta: bool
    Tindex: np.ndarray

    ntrial: int
    naccepted: int

    def __init__(
        self,
        info: odatse.Info,
        runner: odatse.Runner = None,
        domain=None,
        nwalkers: int = 1,
        run_mode: str = "initial",
        mpicomm: Optional["MPI.Comm"] = None,
    ) -> None:
        """
        Initialize the AlgorithmBase class.

        Parameters
        ----------
        info : odatse.Info
            Information object containing algorithm parameters.
        runner : odatse.Runner, optional
            Runner object for executing the algorithm, by default None.
        domain : odatse.domain.Domain, optional
            Domain object defining the problem space, by default None.
        nwalkers : int, optional
            Number of walkers to use in the simulation, by default 1.
        run_mode : str, optional
            Mode of the run, e.g., "initial", by default "initial".
        mpicomm : MPI.Comm
            MPI communicator to use for parallelization.
            If not provided, the default MPI communicator (MPI.COMM_WORLD) will be used if mpi4py is installed.

        Raises
        ------
        ValueError
            If an unsupported domain type is provided or required parameters are missing.

        Examples
        --------
        >>> info = odatse.Info(config_file_path)
        >>> runner = odatse.Runner()
        >>> algorithm = AlgorithmBase(info, runner, nwalkers=100)
        """
        time_sta = time.perf_counter()
        super().__init__(info=info, runner=runner, run_mode=run_mode, mpicomm=mpicomm)
        self.nwalkers = nwalkers

        if domain:
            if isinstance(domain, odatse.domain.MeshGrid):
                self.iscontinuous = False
            elif isinstance(domain, odatse.domain.Region):
                self.iscontinuous = True
            else:
                raise ValueError("ERROR: unsupoorted domain type {}".format(type(domain)))
            self.domain = domain
        else:
            info_param = info.algorithm["param"]
            if "mesh_path" in info_param:
                self.iscontinuous = False
                self.domain = odatse.domain.MeshGrid(info)
            elif "use_grid" in info_param and info_param["use_grid"] == True:
                self.iscontinuous = False
                self.domain = odatse.domain.MeshGrid(info)
            else:
                self.iscontinuous = True
                self.domain = odatse.domain.Region(info)

        if self.iscontinuous:
            self.statespace = ContinuousStateSpace(self.domain, info_param, limitation=self.runner.limitation, rng=self.rng)
        else:
            self.statespace = DiscreteStateSpace(self.domain, info_param, rng=self.rng)

        time_end = time.perf_counter()
        self.timer["init"]["total"] = time_end - time_sta
        self.Tindex = 0
        self.input_as_beta = False

        # -- writer
        self.fp_trial = None
        self.fp_result = None

    def _initialize(self):
        """
        Initialize the Monte Carlo simulation state.

        For continuous problems:
          - Uses domain.initialize to generate valid initial positions
          - Respects any additional limitations from the runner

        For discrete problems:
          - Randomly assigns walkers to valid nodes
          - Maps node indices to actual coordinate positions

        Also initializes:
          - Objective function values (fx) to zero
          - Best solution tracking variables
          - Acceptance counters for monitoring convergence
        """
        self.state = self.statespace.initialize(self.nwalkers)
        self.fx = np.zeros(self.nwalkers)

        self.best_fx = 0.0
        self.best_istep = 0
        self.best_iwalker = 0
        self.naccepted = 0
        self.ntrial = 0

    def _evaluate(self, state, in_range: np.ndarray = None) -> np.ndarray:
        """
        Evaluate objective function for current walker positions.

        Optimization Features:
          - Skips evaluation for out-of-bounds positions
          - Tracks evaluation timing statistics
          - Supports parallel evaluation across walkers

        Parameters
        ----------
        in_range : np.ndarray, optional
            Boolean mask indicating valid positions
            True = position is valid and should be evaluated
            False = position is invalid, will be assigned inf

        Returns
        -------
        np.ndarray
            Array of objective function values
            Invalid positions are assigned inf
        """
        fx = np.zeros(self.nwalkers, dtype=np.float64)
        for iwalker in range(self.nwalkers):
            x = state.x[iwalker, :]
            if in_range is None or in_range[iwalker]:
                args = (self.istep, iwalker)

                time_sta = time.perf_counter()
                fx[iwalker] = self.runner.submit(x, args)
                time_end = time.perf_counter()
                self.timer["run"]["submit"] += time_end - time_sta
            else:
                fx[iwalker] = np.inf
        return fx

    def local_update(
        self,
        beta: Union[float, np.ndarray],
        extra_info_to_write: Union[List, Tuple] = None,
    ) -> None:
        """
        Perform one step of the Monte Carlo algorithm.

        Algorithm Flow:
          1. Generate proposed moves for all walkers
          2. Check if proposals are within valid bounds
          3. Evaluate objective function for valid proposals
          4. Apply Metropolis acceptance criterion:
             ``P(accept) = min(1, exp(-beta * (f_new - f_old)))``
          5. For discrete case, adjust acceptance probability by:
             ``P *= (n_neighbors_old / n_neighbors_new)``
          6. Update positions and energies
          7. Track best solution found
          8. Log results if writers are configured

        Parameters
        ----------
        beta : Union[float, np.ndarray]
            Inverse temperature(s) controlling acceptance probability
            Can be single value or array (one per walker)
        extra_info_to_write : Union[List, Tuple], optional
            Additional data to log with results

        Notes
        -----
          - Handles numerical overflow in exponential calculation
          - Maintains detailed acceptance statistics
          - Supports both single and multiple temperature values
          - Preserves best solution across all steps
        """
        # make candidate
        old_state = copy.deepcopy(self.state)
        old_fx = copy.deepcopy(self.fx)

        new_state, in_range, weight = self.statespace.propose(old_state)
        # self.state = new_state

        # evaluate "Energy"s
        new_fx = self._evaluate(new_state, in_range)
        # XXX
        self.state = new_state
        self.fx = new_fx
        self._write_result(self.fp_trial, extras=extra_info_to_write)

        # print(old_fx, new_fx)
        fdiff = new_fx - old_fx

        # Ignore an overflow warning in np.exp(x) for x >~ 710
        # and an invalid operation warning in exp(nan) (nan came from 0 * inf)
        # Note: fdiff (fx) becomes inf when x is out of range
        # old_setting = np.seterr(over="ignore")
        old_setting = np.seterr(all="ignore")
        probs = np.exp(-beta * fdiff)
        # probs[np.isnan(probs)] = 0.0
        if weight is not None:
            probs *= weight
        np.seterr(**old_setting)

        tocheck = in_range & (probs < 1.0)
        num_check = np.count_nonzero(tocheck)

        accepted = in_range.copy()
        accepted[tocheck] = self.rng.rand(num_check) < probs[tocheck]

        self.naccepted += accepted.sum()
        self.ntrial += accepted.size

        # update
        self.state = self.statespace.choose(accepted, new_state, old_state)
        self.fx = np.where(accepted, new_fx, old_fx)

        minidx = np.argmin(self.fx)
        if self.fx[minidx] < self.best_fx:
            np.copyto(self.best_x, self.state.x[minidx, :])
            self.best_fx = self.fx[minidx]
            self.best_istep = self.istep
            self.best_iwalker = typing.cast(int, minidx)
        self._write_result(self.fp_result, extras=extra_info_to_write)

    def _set_writer(self, fp_trial, fp_result):
        self.fp_trial = fp_trial
        self.fp_result = fp_result

    def _write_result(self, writer, extras=None):
        for iwalker in range(self.nwalkers):
            if isinstance(self.Tindex, int):
                beta = self.betas[self.Tindex]
            else:
                beta = self.betas[self.Tindex[iwalker]]

            if self.input_as_beta:
                tval = beta
            else:
                tval = 1.0 / beta

            data = [self.istep, iwalker, tval, self.fx[iwalker], *self.state.x[iwalker,:]]
            if extras:
                for extra in extras:
                    data.append(extra[iwalker])

            writer.write(*data)



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

from typing import List, Optional, TYPE_CHECKING
import time
import copy
from pathlib import Path
import sys

import physbo
import numpy as np

import odatse
import odatse.domain

if TYPE_CHECKING:
    from mpi4py import MPI

class Algorithm(odatse.algorithm.AlgorithmBase):
    """
    A class to represent the Bayesian optimization algorithm.

    Attributes
    ----------
    mesh_list : np.ndarray
        The mesh grid list.
    label_list : List[str]
        The list of labels.
    random_max_num_probes : int
        The maximum number of random probes.
    bayes_max_num_probes : int
        The maximum number of Bayesian probes.
    score : str
        The scoring method.
    interval : int
        The interval for Bayesian optimization.
    num_rand_basis : int
        The number of random basis.
    xopt : np.ndarray
        The optimal solution.
    best_fx : List[float]
        The list of best function values.
    best_action : List[int]
        The list of best actions.
    fx_list : List[float]
        The list of function values.
    param_list : List[np.ndarray]
        The list of parameters.
    """

    def __init__(
        self,
        info: odatse.Info,
        runner: odatse.Runner = None,
        domain=None,
        run_mode: str = "initial",
        mpicomm: Optional["MPI.Comm"] = None,
    ) -> None:
        """
        Constructs all the necessary attributes for the Algorithm object.

        Parameters
        ----------
        info : odatse.Info
            The information object.
        runner : odatse.Runner, optional
            The runner object (default is None).
        domain : optional
            The domain object (default is None).
        run_mode : str, optional
            The run mode (default is "initial").
        mpicomm : MPI.Comm
            MPI communicator to use for parallelization.
            If not provided, the default MPI communicator (MPI.COMM_WORLD) will be used if mpi4py is installed.
        """
        super().__init__(info=info, runner=runner, run_mode=run_mode, mpicomm=mpicomm)

        info_param = info.algorithm.get("param", {})
        info_bayes = info.algorithm.get("bayes", {})

        for key in ("random_max_num_probes", "bayes_max_num_probes", "score", "interval", "num_rand_basis"):
            if key in info_param and key not in info_bayes:
                print(f"WARNING: algorithm.param.{key} is deprecated. Use algorithm.bayes.{key} .")
                info_bayes[key] = info_param[key]

        self.random_max_num_probes = info_bayes.get("random_max_num_probes", 20)
        self.bayes_max_num_probes = info_bayes.get("bayes_max_num_probes", 40)
        self.score = info_bayes.get("score", "TS")
        self.interval = info_bayes.get("interval", 5)
        self.num_rand_basis = info_bayes.get("num_rand_basis", 5000)

        if self.mpirank == 0:
            print("# parameter")
            print(f"random_max_num_probes = {self.random_max_num_probes}")
            print(f"bayes_max_num_probes = {self.bayes_max_num_probes}")
            print(f"score = {self.score}")
            print(f"interval = {self.interval}")
            print(f"num_rand_basis = {self.num_rand_basis}")

        if domain and isinstance(domain, odatse.domain.MeshGrid):
            self.domain = domain
        else:
            self.domain = odatse.domain.MeshGrid(info, rng=self.rng)
        self.mesh_list = np.array(self.domain.grid)

        X_normalized = physbo.misc.centering(self.mesh_list[:, 1:])
        comm = self.mpicomm if self.mpisize > 1 else None
        if physbo.__version__ < "3":
            self.policy = physbo.search.discrete.policy(test_X=X_normalized, comm=comm)
        else:
            self.policy = physbo.search.discrete.Policy(test_X=X_normalized, comm=comm)

        if "seed" in info.algorithm:
            seed = info.algorithm["seed"]
            self.policy.set_seed(seed)

        self.file_history = "history.npz"
        self.file_training = "training.npz"
        self.file_predictor = "predictor.dump"

    def _initialize(self):
        """
        Initializes the algorithm parameters and timers.
        """
        self.istep = 0
        self.param_list = []
        self.fx_list = []
        self.timer["run"]["random_search"] = 0.0
        self.timer["run"]["bayes_search"] = 0.0

        self._show_parameters()

    def _run(self) -> None:
        """
        Runs the Bayesian optimization process.
        """
        runner = self.runner
        mesh_list = self.mesh_list

        class simulator:
            def __call__(self, action: np.ndarray) -> float:
                """
                Simulates the function evaluation for a given action.

                Parameters
                ----------
                action : np.ndarray
                    The action to be evaluated.

                Returns
                -------
                float
                    The negative function value.
                """
                a = int(action[0])
                args = (a, 0)
                x = mesh_list[a, 1:]
                fx = runner.submit(x, args)
                fx_list.append(fx)
                param_list.append(mesh_list[a])
                return -fx

        if self.mode is None:
            raise RuntimeError("mode unset")

        restore_rng = not self.mode.endswith("-resetrand")

        if self.mode.startswith("init"):
            self._initialize()
        elif self.mode.startswith("resume"):
            self._load_state(self.checkpoint_file, mode="resume", restore_rng=restore_rng)
        elif self.mode.startswith("continue"):
            self._load_state(self.checkpoint_file, mode="continue", restore_rng=restore_rng)
        else:
            raise RuntimeError("unknown mode {}".format(self.mode))

        fx_list = self.fx_list
        param_list = self.param_list

        if self.mode.startswith("init"):
            time_sta = time.perf_counter()
            res = self.policy.random_search(
                max_num_probes=self.random_max_num_probes, simulator=simulator()
            )
            time_end = time.perf_counter()
            self.timer["run"]["random_search"] = time_end - time_sta

            if self.checkpoint:
                self._save_state(self.checkpoint_file)
        else:
            if self.istep >= self.bayes_max_num_probes:
                res = copy.deepcopy(self.policy.history)

        next_checkpoint_step = self.istep + self.checkpoint_steps
        next_checkpoint_time = time.time() + self.checkpoint_interval

        while self.istep < self.bayes_max_num_probes:
            intv = 0 if self.istep % self.interval == 0 else -1

            time_sta = time.perf_counter()
            res = self.policy.bayes_search(
                max_num_probes=1,
                simulator=simulator(),
                score=self.score,
                interval=intv,
                num_rand_basis=self.num_rand_basis,
            )
            time_end = time.perf_counter()
            self.timer["run"]["bayes_search"] += time_end - time_sta

            self.istep += 1

            if self.checkpoint:
                time_now = time.time()
                if self.istep >= next_checkpoint_step or time_now >= next_checkpoint_time:
                    self.fx_list = fx_list
                    self.param_list = param_list

                    self._save_state(self.checkpoint_file)
                    next_checkpoint_step = self.istep + self.checkpoint_steps
                    next_checkpoint_time = time_now + self.checkpoint_interval

        self.best_fx, self.best_action = res.export_all_sequence_best_fx()
        self.xopt = mesh_list[int(self.best_action[-1]), 1:]
        self.fx_list = fx_list
        self.param_list = param_list

        if self.checkpoint:
            self._save_state(self.checkpoint_file)

    def _prepare(self) -> None:
        """
        Prepares the algorithm for execution.
        """
        pass

    def _post(self) -> None:
        """
        Finalizes the algorithm execution and writes the results to a file.
        """
        label_list = self.label_list
        if self.mpirank == 0:
            with open("BayesData.txt", "w") as file_BD:
                file_BD.write("#step")
                for label in label_list:
                    file_BD.write(f" {label}")
                file_BD.write(" fx")
                for label in label_list:
                    file_BD.write(f" {label}_action")
                file_BD.write(" fx_action\n")

                for step, fx in enumerate(self.fx_list):
                    file_BD.write(str(step))
                    best_idx = int(self.best_action[step])
                    for v in self.mesh_list[best_idx][1:]:
                        file_BD.write(f" {v}")
                    file_BD.write(f" {-self.best_fx[step]}")

                    for v in self.param_list[step][1:]:
                        file_BD.write(f" {v}")
                    file_BD.write(f" {fx}\n")
            print("Best Solution:")
            for x, y in zip(label_list, self.xopt):
                print(x, "=", y)
        return {"x": self.xopt, "fx": self.best_fx}

    def _save_state(self, filename):
        """
        Saves the current state of the algorithm to a file.

        Parameters
        ----------
        filename : str
            The name of the file to save the state.
        """
        data = {
            "mpisize": self.mpisize,
            "mpirank": self.mpirank,
            "rng": self.rng.get_state(),
            "timer": self.timer,
            "info": self.info,
            "istep": self.istep,
            "param_list": self.param_list,
            "fx_list": self.fx_list,
            "file_history": self.file_history,
            "file_training": self.file_training,
            "file_predictor": self.file_predictor,
            "random_number": np.random.get_state(),
        }
        self._save_data(data, filename)

        self.policy.save(file_history=Path(self.output_dir, self.file_history),
                         file_training=Path(self.output_dir, self.file_training),
                         file_predictor=Path(self.output_dir, self.file_predictor))

    def _load_state(self, filename, mode="resume", restore_rng=True):
        """
        Loads the state of the algorithm from a file.

        Parameters
        ----------
        filename : str
            The name of the file to load the state from.
        mode : str, optional
            The mode to load the state (default is "resume").
        restore_rng : bool, optional
            Whether to restore the random number generator state (default is True).
        """
        data = self._load_data(filename)
        if not data:
            print("ERROR: Load status file failed")
            sys.exit(1)

        assert self.mpisize == data["mpisize"]
        assert self.mpirank == data["mpirank"]

        if restore_rng:
            self.rng = np.random.RandomState()
            self.rng.set_state(data["rng"])
            np.random.set_state(data["random_number"])
        self.timer = data["timer"]

        info = data["info"]
        self._check_parameters(info)

        self.istep = data["istep"]
        self.param_list = data["param_list"]
        self.fx_list = data["fx_list"]

        self.policy.load(file_history=Path(self.output_dir, self.file_history),
                         file_training=Path(self.output_dir, self.file_training),
                         file_predictor=Path(self.output_dir, self.file_predictor))



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

from abc import ABCMeta, abstractmethod

import subprocess
import numpy as np

import odatse
import odatse.mpi

# type hints
from pathlib import Path
from typing import Dict, List, Tuple


class SolverBase(object, metaclass=ABCMeta):
    """
    Abstract base class for solvers in the 2DMAT software.
    """

    root_dir: Path
    output_dir: Path
    proc_dir: Path
    work_dir: Path
    _name: str
    dimension: int
    timer: Dict[str, Dict]

    @abstractmethod
    def __init__(self, info: odatse.Info) -> None:
        """
        Initialize the solver with the given information.

        Parameters
        ----------
        info : Info
            Information object containing configuration details.
        """
        self.root_dir = info.base["root_dir"]
        self.output_dir = info.base["output_dir"]
        self.proc_dir = self.output_dir / str(odatse.mpi.rank())
        self.work_dir = self.proc_dir
        self._name = ""
        self.timer = {"prepare": {}, "run": {}, "post": {}}
        if "dimension" in info.solver:
            self.dimension = info.solver["dimension"]
        else:
            self.dimension = info.base["dimension"]

    @property
    def name(self) -> str:
        """
        Get the name of the solver.

        Returns
        -------
        str
            The name of the solver.
        """
        return self._name

    @abstractmethod
    def evaluate(self, x: np.ndarray, arg: Tuple = (), nprocs: int = 1, nthreads: int = 1) -> None:
        """
        Evaluate the solver with the given parameters.

        Parameters
        ----------
        x : np.ndarray
            Input data array.
        arg : Tuple, optional
            Additional arguments for evaluation. Defaults to ().
        nprocs : nt, optional
            Number of processes to use. Defaults to 1.
        nthreads : int, optional
            Number of threads to use. Defaults to 1.

        Raises
        ------
        NotImplementedError
            This method should be implemented by subclasses.
        """
        raise NotImplementedError()



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

from typing import Union, List

import numpy as np


def read_vector(inp: Union[str, List[float]]) -> np.ndarray:
    """
    Converts an input string or list of floats into a numpy array vector.

    Parameters
    ----------
    inp : Union[str, List[float]]
        Input data, either as a space-separated string of numbers or a list of floats.

    Returns
    -------
    np.ndarray
        A numpy array representing the vector.

    Raises
    ------
    RuntimeError
        If the input is not a vector.
    """
    if isinstance(inp, str):
        vlist = [float(w) for w in inp.split()]
    else:
        vlist = inp
    v = np.array(vlist)
    if v.ndim > 1:
        msg = f"input is not vector ({inp})"
        raise RuntimeError(msg)
    return v

def read_matrix(inp: Union[str, List[List[float]]]) -> np.ndarray:
    """
    Converts an input string or list of lists of floats into a numpy array matrix.

    Parameters
    ----------
    inp : Union[str, List[List[float]]]
        Input data, either as a string with rows of space-separated numbers or a list of lists of floats.

    Returns
    -------
    np.ndarray
        A numpy array representing the matrix.

    Raises
    ------
    RuntimeError
        If the input is not a matrix.
    """
    if isinstance(inp, str):
        Alist: List[List[float]] = []
        for line in inp.split("\n"):
            if not line.strip():  # empty
                continue
            Alist.append([float(w) for w in line.strip().split()])
    else:
        Alist = inp
    A = np.array(Alist)
    if A.size == 0 or A.ndim == 2:
        return A
    msg = f"input is not matrix ({inp})"
    raise RuntimeError(msg)



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import sys
import odatse

def main():
    """
    Main function to run the data-analysis software for quantum beam diffraction experiments
    on 2D material structures. It parses command-line arguments, loads the input file,
    selects the appropriate algorithm and solver, and executes the analysis.
    """

    info, run_mode = odatse.initialize()

    alg_module = odatse.algorithm.choose_algorithm(info.algorithm["name"])

    solvername = info.solver["name"]
    if solvername == "analytical":
        from .solver.analytical import Solver
    else:
        if odatse.mpi.rank() == 0:
            print(f"ERROR: Unknown solver ({solvername})")
        sys.exit(1)

    solver = Solver(info)
    runner = odatse.Runner(solver, info)
    alg = alg_module.Algorithm(info, runner, run_mode=run_mode)

    result = alg.main()



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

from typing import List, Dict, Union, Any

from pathlib import Path
import numpy as np

import odatse

class DomainBase:
    """
    Base class for domain management in the 2DMAT software.

    Attributes
    ----------
    root_dir : Path
        The root directory for the domain.
    output_dir : Path
        The output directory for the domain.
    mpisize : int
        The size of the MPI communicator.
    mpirank : int
        The rank of the MPI process.
    """
    def __init__(self, info: odatse.Info = None):
        """
        Initializes the DomainBase instance.

        Parameters
        ----------
        info : Info, optional
            An instance of odatse.Info containing base directory information.
        """
        if info:
            self.root_dir = info.base["root_dir"]
            self.output_dir = info.base["output_dir"]
        else:
            self.root_dir = Path(".")
            self.output_dir = Path(".")

        self.mpisize = odatse.mpi.size()
        self.mpirank = odatse.mpi.rank()



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

from typing import List, Dict, Union, Optional, TYPE_CHECKING

from io import open
import copy
import time
import sys

import numpy as np

import odatse
import odatse.exception
import odatse.algorithm.montecarlo
from odatse.algorithm.gather import gather_replica, gather_data
import odatse.util.resampling
from odatse.util.read_ts import read_Ts
from odatse.util.data_writer import DataWriter

if TYPE_CHECKING:
    from mpi4py import MPI


class Algorithm(odatse.algorithm.montecarlo.AlgorithmBase):
    """
    Population Annealing Monte Carlo (PAMC) Algorithm Implementation.
    
    PAMC is an advanced Monte Carlo method that combines features of simulated annealing
    and population-based methods. It maintains a population of walkers that evolves
    through a sequence of temperatures, using both Monte Carlo updates and resampling.

    Key Features:
      - Maintains a population of walkers that evolves through temperature steps
      - Uses importance sampling with weight-based resampling
      - Supports both fixed and variable population sizes
      - Calculates free energy differences between temperature steps
      - Provides error estimates through population statistics

    Algorithm Flow:
      1. Initialize walker population at highest temperature
      2. For each temperature step:
        a. Perform Monte Carlo updates at current temperature
        b. Calculate weights for next temperature
        c. Resample population based on weights
        d. Update statistical estimates
      3. Track best solutions and maintain system statistics

    Attributes
    ----------
    x : np.ndarray
        Current configurations for all walkers
    logweights : np.ndarray
        Log of importance weights for each walker
    fx : np.ndarray
        Current energy/objective values
    nreplicas : np.ndarray
        Number of replicas at each temperature
    betas : np.ndarray
        Inverse temperatures (β = 1/T)
    Tindex : int
        Current temperature index
    Fmeans : np.ndarray
        Mean free energy at each temperature
    Ferrs : np.ndarray
        Free energy error estimates
    walker_ancestors : np.ndarray
        Tracks genealogy of walkers for analysis
    """

    # x: np.ndarray
    # xmin: np.ndarray
    # xmax: np.ndarray
    # #xunit: np.ndarray
    # xstep: np.ndarray

    numsteps: int
    numsteps_annealing: int

    logweights: np.ndarray
    fx: np.ndarray
    istep: int
    nreplicas: np.ndarray
    betas: np.ndarray
    Tindex: int

    fx_from_reset: np.ndarray

    # 0th column: number of accepted MC trials
    # 1st column: number of total MC trials
    naccepted_from_reset: np.ndarray
    resampling_interval: int

    nset_bootstrap: int

    walker_ancestors: np.ndarray
    populations: np.ndarray
    family_lo: int
    family_hi: int

    Fmeans: np.ndarray
    Ferrs: np.ndarray

    def __init__(
        self,
        info: odatse.Info,
        runner: odatse.Runner = None,
        run_mode: str = "initial",
        mpicomm: Optional["MPI.Comm"] = None,
    ) -> None:
        """
        Initialize the Algorithm class.

        Parameters
        ----------
        info : odatse.Info
            Information object containing algorithm parameters.
        runner : odatse.Runner, optional
            Runner object for executing the algorithm, by default None.
        run_mode : str, optional
            Mode in which to run the algorithm, by default "initial".
        mpicomm : MPI.Comm
            MPI communicator to use for parallelization.
            If not provided, the default MPI communicator (MPI.COMM_WORLD) will be used if mpi4py is installed.
        """
        time_sta = time.perf_counter()

        info_pamc = info.algorithm["pamc"]
        nwalkers = info_pamc.get("nreplica_per_proc", 1)

        super().__init__(info=info, runner=runner, nwalkers=nwalkers, run_mode=run_mode, mpicomm=mpicomm)

        self.verbose = True and self.mpirank == 0

        numT = self._find_scheduling(info_pamc)

        self.input_as_beta, self.betas = read_Ts(info_pamc, numT=numT)
        self.betas.sort()

        self.fix_nwalkers = info_pamc.get("fix_num_replicas", True)
        self.resampling_interval = info_pamc.get("resampling_interval", 1)
        if self.resampling_interval < 1:
            self.resampling_interval = numT + 1

        self.export_combined_files = info_pamc.get("export_combined_files", False)
        self.separate_T = info_pamc.get("separate_T", True)

        time_end = time.perf_counter()
        self.timer["init"]["total"] = time_end - time_sta

    def _initialize(self) -> None:
        super()._initialize()

        numT = len(self.betas)

        self.logZ = 0.0
        self.logZs = np.zeros(numT)
        self.logweights = np.zeros(self.nwalkers)

        self.Fmeans = np.zeros(numT)
        self.Ferrs = np.zeros(numT)
        nreplicas = self.mpisize * self.nwalkers
        self.nreplicas = np.full(numT, nreplicas)

        self.populations = np.zeros((numT, self.nwalkers), dtype=int)
        self.family_lo = self.nwalkers * self.mpirank
        self.family_hi = self.nwalkers * (self.mpirank + 1)
        self.walker_ancestors = np.arange(self.family_lo, self.family_hi)
        self.fx_from_reset = np.zeros((self.resampling_interval, self.nwalkers))
        self.naccepted_from_reset = np.zeros((self.resampling_interval, 2), dtype=int)
        self.acceptance_ratio = np.zeros(numT)
        self.pr_list = np.zeros(numT)

        self._show_parameters()

    def _find_scheduling(self, info_pamc) -> int:
        """
        Determine the temperature schedule and number of steps.

        The schedule can be specified in three ways:
          1. Total steps and steps per temperature
          2. Total steps and number of temperatures
          3. Steps per temperature and number of temperatures

        The method ensures even distribution of computational effort across
        temperature steps while respecting the specified constraints.

        Parameters
        ----------
        info_pamc : dict
            Configuration dictionary containing:
              - numsteps: Total number of Monte Carlo steps
              - numsteps_annealing: Steps per temperature
              - Tnum: Number of temperature points

        Returns
        -------
        int
            Number of temperature steps in the schedule

        Raises
        ------
        odatse.exception.InputError
            If the scheduling parameters are inconsistent
        """
        numsteps = info_pamc.get("numsteps", 0)
        numsteps_annealing = info_pamc.get("numsteps_annealing", 0)
        numT = info_pamc.get("Tnum", 0)

        oks = np.array([numsteps, numsteps_annealing, numT]) > 0
        if np.count_nonzero(oks) != 2:
            msg = "ERROR: Two of 'numsteps', 'numsteps_annealing', "
            msg += "and 'Tnum' should be positive in the input file\n"
            msg += f"  numsteps = {numsteps}\n"
            msg += f"  numsteps_annealing = {numsteps_annealing}\n"
            msg += f"  Tnum = {numT}\n"
            raise odatse.exception.InputError(msg)

        if numsteps <= 0:
            self.numsteps_for_T = np.full(numT, numsteps_annealing)
        elif numsteps_annealing <= 0:
            nr = numsteps // numT
            self.numsteps_for_T = np.full(numT, nr)
            rem = numsteps - nr * numT
            if rem > 0:
                self.numsteps_for_T[0 : (rem - 1)] += 1
        else:
            ss: List[int] = []
            while numsteps > 0:
                if numsteps > numsteps_annealing:
                    ss.append(numsteps_annealing)
                    numsteps -= numsteps_annealing
                else:
                    ss.append(numsteps)
                    numsteps = 0
            self.numsteps_for_T = np.array(ss)
            numT = len(ss)

        return numT

    def _run(self) -> None:

        if self.mode is None:
            raise RuntimeError("mode unset")

        restore_rng = not self.mode.endswith("-resetrand")

        if self.mode.startswith("init"):
            self._initialize()

            self.Tindex = 0
            self.index_from_reset = 0
            self.istep = 0

        elif self.mode.startswith("resume"):
            self._load_state(self.checkpoint_file, mode="resume", restore_rng=restore_rng)

        elif self.mode.startswith("continue"):
            self._load_state(self.checkpoint_file, mode="continue", restore_rng=restore_rng)

            Tindex = self.Tindex

            dbeta = self.betas[Tindex + 1] - self.betas[Tindex]
            self.logweights += -dbeta * self.fx
            if self.index_from_reset == self.resampling_interval:
                time_sta = time.perf_counter()
                self._resample()
                time_end = time.perf_counter()
                self.timer["run"]["resampling"] += time_end - time_sta
                self.index_from_reset = 0

            self.Tindex += 1
        else:
            raise RuntimeError("unknown mode {}".format(self.mode))

        writer = self._setup_writer()

        if self.mode.startswith("init"):
            beta = self.betas[self.Tindex]
            self.fx = self._evaluate(self.state)

            self._write_result(writer["trial"], [np.exp(self.logweights), self.walker_ancestors])
            self._write_result(writer["result"], [np.exp(self.logweights), self.walker_ancestors])

            self.istep += 1

            minidx = np.argmin(self.fx)
            self.best_x = copy.copy(self.state.x[minidx, :])
            self.best_fx = np.min(self.fx[minidx])
            self.best_istep = 0
            self.best_iwalker = 0

        next_checkpoint_step = self.istep + self.checkpoint_steps
        next_checkpoint_time = time.time() + self.checkpoint_interval

        if self.verbose:
            print("\u03b2 mean[f] Err[f] nreplica log(Z/Z0) acceptance_ratio")

        numT = len(self.betas)
        while self.Tindex < numT:
            # print(">>> Tindex = {}".format(self.Tindex))
            Tindex = self.Tindex
            beta = self.betas[self.Tindex]

            if self.nwalkers != 0:
                for _ in range(self.numsteps_for_T[Tindex]):
                    self.local_update(
                        beta,
                        extra_info_to_write=[
                            np.exp(self.logweights),
                            self.walker_ancestors,
                        ],
                    )
                    self.istep += 1
            else : #self.nwalkers == 0
                pass

            # print(">>> istep={}".format(self.istep))

            self.fx_from_reset[self.index_from_reset, :] = self.fx[:]
            self.naccepted_from_reset[self.index_from_reset, 0] = self.naccepted
            self.naccepted_from_reset[self.index_from_reset, 1] = self.ntrial
            self.naccepted = 0
            self.ntrial = 0
            self.index_from_reset += 1

            # write weights
            if writer["weight"]:
                for iwalker in range(self.nwalkers):
                    writer["weight"].write(Tindex,
                                           beta,
                                           iwalker,
                                           0,
                                           self.fx[iwalker],
                                           self.logweights[iwalker],
                                           *self.state.x[iwalker,:])

            # calculate participation ratio
            pr = self._calc_participation_ratio()
            self.pr_list[Tindex] = pr

            if Tindex == numT - 1:
                break

            dbeta = self.betas[Tindex + 1] - self.betas[Tindex]
            self.logweights += -dbeta * self.fx
            if self.index_from_reset == self.resampling_interval:
                time_sta = time.perf_counter()
                self._resample()
                time_end = time.perf_counter()
                self.timer["run"]["resampling"] += time_end - time_sta
                self.index_from_reset = 0

            self.Tindex += 1

            if self.checkpoint:
                time_now = time.time()
                if self.istep >= next_checkpoint_step or time_now >= next_checkpoint_time:
                    print(">>> checkpoint")
                    self._save_state(self.checkpoint_file)
                    next_checkpoint_step = self.istep + self.checkpoint_steps
                    next_checkpoint_time = time_now + self.checkpoint_interval

        # store final state for continuation
        if self.checkpoint:
            print(">>> store final state")
            self._save_state(self.checkpoint_file)

        if self.index_from_reset > 0:
            res = self._gather_information(self.index_from_reset)
            self._save_stats(res)

        # must close explicitly
        self._close_writer(writer)

        if self.separate_T and not self.export_combined_files:
            self._split_result_file("trial")
            self._split_result_file("result")

        if self.mpisize > 1:
            self.mpicomm.barrier()
        print("complete main process : rank {:08d}/{:08d}".format(self.mpirank, self.mpisize))

    def _setup_writer(self):
        write_mode = "w" if self.mode.startswith("init") else "a"

        item_list = [
            "step",
            "walker",
            ("beta" if self.input_as_beta else "T"),
            "fx",
            *self.label_list,
            "weight",
            "ancestor",
        ]

        file_trial = DataWriter("trial.txt", mode=write_mode, item_list=item_list, combined=self.export_combined_files)
        file_result = DataWriter("result.txt", mode=write_mode, item_list=item_list, combined=self.export_combined_files)
        self._set_writer(file_trial, file_result)

        item_list_weight = [
            "Tindex",
            "beta",
            "walker",
            "idnum",
            "fx",
            "log_weight",
            *self.label_list,
        ]
        file_weight = DataWriter("weight.txt", mode=write_mode, item_list=item_list_weight, combined=self.export_combined_files)

        return { "trial": file_trial, "result": file_result, "weight": file_weight }

    def _close_writer(self, writers):
        for k, v in writers.items():
            if v:
                v.close()

    def _gather_information(self, numT: int = None) -> Dict[str, np.ndarray]:
        """
        Collect and organize statistical information across all processes.

        Gathers data needed for:
          1. Free energy calculations
          2. Error estimation
          3. Population statistics
          4. Acceptance rate monitoring

        Parameters
        ----------
        numT : int, optional
            Number of temperature steps to gather data for

        Returns
        -------
        Dict[str, np.ndarray]
            Contains:
              - fxs: Energy values for all walkers
              - logweights: Log of importance weights
              - ns: Number of walkers per process
              - ancestors: Genealogical tracking data
              - acceptance ratio: MC acceptance rates
        """

        if numT is None:
            numT = self.resampling_interval

        res = {}

        res["ns"] = gather_data(np.array([self.nwalkers]))
        res["fxs"] = gather_replica(self.fx_from_reset[0:numT,:], axis=1)
        res["ancestors"] = gather_replica(self.walker_ancestors)
        nacc = gather_data([self.naccepted_from_reset[0:numT,:]])
        nacc = np.sum(nacc, axis=0)
        res["acceptance ratio"] = nacc[:,0] / nacc[:,1]

        fxs = res["fxs"]
        nreplicas = np.sum(res["ns"])

        betas = self.betas[self.Tindex-numT+1 : self.Tindex+1]
        lw = np.zeros((numT, nreplicas), dtype=np.float64)
        for i in range(1, numT):
            dbeta = betas[i] - betas[i-1]
            lw[i, :] = lw[i-1, :] - dbeta * fxs[i-1, :]
        res["logweights"] = lw

        return res

    def _save_stats(self, info: Dict[str, np.ndarray]) -> None:
        """
        Calculate and save statistical measures from the simulation.

        Performs:
          1. Free energy calculations using weighted averages
          2. Error estimation using jackknife resampling
          3. Population size tracking
          4. Acceptance rate monitoring
          5. Partition function estimation

        Uses bias-corrected jackknife for reliable error estimates
        of weighted averages in the presence of correlations.

        Parameters
        ----------
        info : Dict[str, np.ndarray]
            Dictionary containing the following keys:
              - fxs: Objective function of each walker over all processes.
              - logweights: Logarithm of weights.
              - ns: Number of walkers in each process.
              - ancestors: Ancestor (origin) of each walker.
              - acceptance ratio: Acceptance ratio for each temperature.
        """
        fxs = info["fxs"]
        numT, nreplicas = fxs.shape
        endTindex = self.Tindex + 1
        startTindex = endTindex - numT

        logweights = info["logweights"]
        logweights_max = logweights.max(axis=1).reshape(-1, 1)
        weights = np.exp(logweights - logweights_max)  # to avoid overflow

        # Bias-corrected jackknife resampling method
        fs = np.zeros((numT, nreplicas))
        fw_sum = (fxs * weights).sum(axis=1)
        w_sum = weights.sum(axis=1)
        for i in range(nreplicas):
            F = fw_sum - fxs[:, i] * weights[:, i]
            W = w_sum - weights[:, i]
            fs[:, i] = F / W
        N = fs.shape[1]
        fm = N * (fw_sum / w_sum) - (N - 1) * fs.mean(axis=1)
        ferr = np.sqrt((N - 1) * fs.var(axis=1))

        self.Fmeans[startTindex:endTindex] = fm
        self.Ferrs[startTindex:endTindex] = ferr
        self.nreplicas[startTindex:endTindex] = nreplicas
        self.acceptance_ratio[startTindex:endTindex] = info["acceptance ratio"][0:numT]

        logz = np.log(np.mean(weights, axis=1))
        logz += logweights_max.flatten()
        self.logZs[startTindex:endTindex] = self.logZ + logz
        if endTindex < len(self.betas):
            # Calculate the next weight before reset and evaluate dF
            bdiff = self.betas[endTindex] - self.betas[endTindex - 1]
            w = np.exp(logweights[-1, :] - bdiff * fxs[-1, :])
            self.logZ = self.logZs[startTindex] + np.log(w.mean())

        if self.verbose:
            for iT in range(startTindex, endTindex):
                print(" ".join(map(str, [
                    self.betas[iT],
                    self.Fmeans[iT],
                    self.Ferrs[iT],
                    self.nreplicas[iT],
                    self.logZs[iT],
                    self.acceptance_ratio[iT],
                ])))

    def _resample(self) -> None:
        """
        Perform population resampling between temperature steps.

        This is a key component of PAMC that:
          1. Gathers current population statistics
          2. Calculates importance weights for the temperature change
          3. Resamples walkers based on their weights
          4. Updates population statistics and free energy estimates

        The resampling can be done in two modes:
          - Fixed: Maintains constant population size
          - Varied: Allows population size to fluctuate based on weights

        Implementation Details:
          - Uses log-weights to prevent numerical overflow
          - Maintains walker genealogy for analysis
          - Updates free energy estimates using resampling data
          - Handles MPI communication for parallel execution
        """
        res = self._gather_information()
        self._save_stats(res)

        # weights for resampling
        dbeta = self.betas[self.Tindex + 1] - self.betas[self.Tindex]
        logweights = res["logweights"][-1, :] - dbeta * res["fxs"][-1, :]
        weights = np.exp(logweights - logweights.max())  # to avoid overflow
        if self.fix_nwalkers:
            self._resample_fixed(weights)
            self.logweights[:] = 0.0
        else:
            ns = res["ns"]
            offsets = np.cumsum(ns) - ns
            self._resample_varied(weights, offsets[self.mpirank])
            self.fx_from_reset = np.zeros((self.resampling_interval, self.nwalkers))
            self.logweights = np.zeros(self.nwalkers)

    def _resample_fixed(self, weights: np.ndarray) -> None:
        """
        Perform resampling with fixed weights.

        This method resamples the walkers based on the provided weights and updates
        the state of the algorithm accordingly.

        Parameters
        ----------
        weights : np.ndarray
            Array of weights for resampling.
        """
        resampler = odatse.util.resampling.WalkerTable(weights)
        new_index = resampler.sample(self.rng, self.nwalkers)

        states = self.statespace.gather(self.state)
        ancestors = gather_replica(self.walker_ancestors)

        self.state = self.statespace.pick(states, new_index)
        self.walker_ancestors = ancestors[new_index]

        fxs = gather_replica(self.fx)
        self.fx = fxs[new_index]

    def _resample_varied(self, weights: np.ndarray, offset: int) -> None:
        """
        Resample population allowing size variation.

        Uses Poisson resampling:
          1. Calculates expected number of copies from weights
          2. Samples actual copies using Poisson distribution
          3. Creates new population with variable size
          4. Updates all walker properties

        Parameters
        ----------
        weights : np.ndarray
            Importance weights for resampling
        offset : int
            Process-specific offset in global population
        """
        weights_sum = np.sum(weights)
        expected_numbers = (self.nreplicas[0] / weights_sum) * weights[
            offset : offset + self.nwalkers
        ]
        next_numbers = self.rng.poisson(expected_numbers)

        new_index = []
        for iwalker, num in enumerate(next_numbers):
            new_index.extend([iwalker] * num)

        new_state = self.statespace.pick(self.state, new_index)
        new_fx = self.fx[new_index]

        self.state = new_state
        self.fx = new_fx
        self.walker_ancestors = np.array(new_index)

        self.nwalkers = np.sum(next_numbers)

    def _calc_participation_ratio(self) -> float:
        """
        Calculate the participation ratio of the current walker population.
        
        The participation ratio is a measure of the effective sample size and
        indicates how evenly distributed the weights are among walkers. It is
        calculated as (sum(w))²/sum(w²), where w are the normalized weights.
        
        A value close to the total number of walkers indicates well-distributed weights,
        while a small value indicates that only a few walkers dominate the population.
        
        To avoid numerical issues with large log-weights, we normalize by subtracting
        the maximum log-weight before exponentiation.
        
        Parameters
        ----------
        None
            Uses the current state of self.logweights
            
        Returns
        -------
        float
            The participation ratio, a value between 1 and the total number of walkers
            
        Notes
        -----
        In parallel execution, this method aggregates weights across all processes
        to calculate the global participation ratio.
        """
        log_weights = gather_replica(self.logweights)
        max_log_weight = np.max(log_weights)

        sum_weight = np.sum(np.exp(log_weights - max_log_weight))
        sum_weight_sq = np.sum(np.exp(log_weights - max_log_weight)**2)

        pr = sum_weight ** 2 / sum_weight_sq

        return pr

    def _prepare(self) -> None:
        """
        Prepare the algorithm for execution.

        This method initializes the timers for the 'submit' and 'resampling' phases
        of the algorithm run.
        """
        self.timer["run"]["submit"] = 0.0
        self.timer["run"]["resampling"] = 0.0

    def _split_result_file(self, tag):
        current_beta = -1
        header_str = ""
        idx = 0
        fwrite = None

        with open(tag + ".txt", "r") as fread:
            for line in fread:
                if line.startswith("#"):
                    header_str += line
                else:
                    data = line.split()
                    beta = data[2]
                    if beta != current_beta:
                        if fwrite:
                            fwrite.close()
                            idx += 1
                        fwrite = open(tag + f"_T{idx}.txt", "w")
                        fwrite.write(header_str)
                        current_beta = beta
                    fwrite.write(line)
            if fwrite:
                fwrite.close()

    def _post(self) -> Dict:
        """
        Post-processing after the algorithm execution.

        This method consolidates the results from different temperature steps
        into single files for 'result' and 'trial'. It also gathers the best
        results from all processes and writes them to 'best_result.txt'.
        """
        best_fx = gather_data(np.array([self.best_fx]))
        best_x = gather_data(np.array([self.best_x]))
        best_istep = gather_data(np.array([self.best_istep]))
        best_iwalker = gather_data(np.array([self.best_iwalker]))

        best_rank = np.argmin(best_fx)
        if self.mpirank == 0:
            with open("best_result.txt", "w") as f:
                f.write(f"nprocs = {self.mpisize}\n")
                f.write(f"rank = {best_rank}\n")
                f.write(f"step = {best_istep[best_rank]}\n")
                f.write(f"walker = {best_iwalker[best_rank]}\n")
                f.write(f"fx = {best_fx[best_rank]}\n")
                for label, x in zip(self.label_list, best_x[best_rank]):
                    f.write(f"{label} = {x}\n")
            print("Best Result:")
            print(f"  rank = {best_rank}")
            print(f"  step = {best_istep[best_rank]}")
            print(f"  walker = {best_iwalker[best_rank]}")
            print(f"  fx = {best_fx[best_rank]}")
            for label, x in zip(self.label_list, best_x[best_rank]):
                print(f"  {label} = {x}")

            with open("fx.txt", "w") as f:
                f.write("# $1: 1/T\n")
                f.write("# $2: mean of f(x)\n")
                f.write("# $3: standard error of f(x)\n")
                f.write("# $4: number of replicas\n")
                f.write("# $5: log(Z/Z0)\n")
                f.write("# $6: acceptance ratio\n")
                for i in range(len(self.betas)):
                    f.write(f"{self.betas[i]}")
                    f.write(f" {self.Fmeans[i]} {self.Ferrs[i]}")
                    f.write(f" {self.nreplicas[i]}")
                    f.write(f" {self.logZs[i]}")
                    f.write(f" {self.acceptance_ratio[i]}")
                    f.write("\n")

            with open("pr.txt", "w") as f:
                f.write("# $1: Tindex\n")
                f.write("# $2: 1/T\n")
                f.write("# $3: participation ratio\n")
                for i in range(len(self.betas)):
                    f.write(f"{i}")
                    f.write(f" {self.betas[i]}")
                    f.write(f" {self.pr_list[i]}")
                    f.write("\n")

        return {
            "x": best_x[best_rank],
            "fx": best_fx[best_rank],
            "nprocs": self.mpisize,
            "rank": best_rank,
            "step": best_istep[best_rank],
            "walker": best_iwalker[best_rank],
        }

    def _save_state(self, filename) -> None:
        """
        Save the current state of the algorithm to a file.

        Parameters
        ----------
        filename : str
            The name of the file where the state will be saved.
        """
        data = {
            #-- _algorithm
            "mpisize": self.mpisize,
            "mpirank": self.mpirank,
            "rng": self.rng.get_state(),
            "timer": self.timer,
            "info": self.info,
            #-- montecarlo
            "x": self.state,
            "fx": self.fx,
            #"inode": self.inode,
            "istep": self.istep,
            "best_x": self.best_x,
            "best_fx": self.best_fx,
            "best_istep": self.best_istep,
            "best_iwalker": self.best_iwalker,
            "naccepted": self.naccepted,
            "ntrial": self.ntrial,
            #-- pamc
            "betas": self.betas,
            "nwalkers": self.nwalkers,
            "input_as_beta": self.input_as_beta,
            "numsteps_for_T": self.numsteps_for_T,
            "Tindex": self.Tindex,
            "index_from_reset": self.index_from_reset,
            "logZ": self.logZ,
            "logZs": self.logZs,
            "logweights": self.logweights,
            "Fmeans": self.Fmeans,
            "Ferrs": self.Ferrs,
            "nreplicas": self.nreplicas,
            "populations": self.populations,
            "family_lo": self.family_lo,
            "family_hi": self.family_hi,
            "walker_ancestors": self.walker_ancestors,
            "fx_from_reset": self.fx_from_reset,
            "naccepted_from_reset": self.naccepted_from_reset,
            "acceptance_ratio": self.acceptance_ratio,
            "pr_list": self.pr_list,
        }
        self._save_data(data, filename)

    def _load_state(self, filename, mode="resume", restore_rng=True):
        """
        Load the saved state of the algorithm from a file.

        Parameters
        ----------
        filename : str
            The name of the file from which the state will be loaded.
        mode : str, optional
            The mode in which to load the state. Can be "resume" or "continue", by default "resume".
        restore_rng : bool, optional
            Whether to restore the random number generator state, by default True.
        """
        data = self._load_data(filename)
        if not data:
            print("ERROR: Load status file failed")
            sys.exit(1)

        # -- _algorithm
        assert self.mpisize == data["mpisize"]
        assert self.mpirank == data["mpirank"]

        if restore_rng:
            self.rng = np.random.RandomState()
            self.rng.set_state(data["rng"])
        self.timer = data["timer"]

        info = data["info"]
        self._check_parameters(info)

        # -- montecarlo
        self.state = data["x"]
        self.fx = data["fx"]
        # self.inode = data["inode"]

        self.istep = data["istep"]

        self.best_x = data["best_x"]
        self.best_fx = data["best_fx"]
        self.best_istep = data["best_istep"]
        self.best_iwalker = data["best_iwalker"]

        self.naccepted = data["naccepted"]
        self.ntrial = data["ntrial"]

        # -- pamc
        self.Tindex = data["Tindex"]
        self.index_from_reset = data["index_from_reset"]
        self.nwalkers = data["nwalkers"]

        if mode == "resume":
            # check if scheduling is as stored
            betas = data["betas"]
            input_as_beta = data["input_as_beta"]
            numsteps_for_T = data["numsteps_for_T"]

            assert np.all(betas == self.betas)
            assert input_as_beta == self.input_as_beta
            assert np.all(numsteps_for_T == self.numsteps_for_T)
            assert self.Tindex < len(self.betas)

        elif mode == "continue":
            # check if scheduling is continuous
            betas = data["betas"]
            input_as_beta = data["input_as_beta"]
            numsteps_for_T = data["numsteps_for_T"]

            assert input_as_beta == self.input_as_beta
            if not betas[-1] == self.betas[0]:
                print("ERROR: temperator is not continuous")
                sys.exit(1)
            self.betas = np.concatenate([betas, self.betas[1:]])
            self.numsteps_for_T = np.concatenate([numsteps_for_T, self.numsteps_for_T[1:]])

        else:
            pass

        numT = len(self.betas)

        nreplicas = self.mpisize * self.nwalkers

        self.logZs = np.zeros(numT)
        self.Fmeans = np.zeros(numT)
        self.Ferrs = np.zeros(numT)
        self.nreplicas = np.full(numT, nreplicas)
        self.acceptance_ratio = np.zeros(numT)
        self.pr_list = np.zeros(numT)

        self.logZ = data["logZ"]
        self.logZs[0:len(data["logZs"])] = data["logZs"]
        self.logweights = data["logweights"]
        self.Fmeans[0:len(data["Fmeans"])] = data["Fmeans"]
        self.Ferrs[0:len(data["Ferrs"])] = data["Ferrs"]
        self.nreplicas[0:len(data["nreplicas"])] = data["nreplicas"]
        self.populations = data["populations"].copy()

        self.family_lo = data["family_lo"]
        self.family_hi = data["family_hi"]

        self.fx_from_reset = data["fx_from_reset"]
        self.walker_ancestors = data["walker_ancestors"]
        self.naccepted_from_reset = data["naccepted_from_reset"]
        self.acceptance_ratio[0:len(data["acceptance_ratio"])] = data["acceptance_ratio"]
        self.pr_list[0:len(data["pr_list"])] = data["pr_list"]

        # -- restore rng state in statespace
        self.statespace.rng = self.rng



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import logging

def basicConfig(combined_filename=None, combined_mode=None):
    return BaseWriter.basicConfig(combined_filename, combined_mode)

class BaseWriter:
    """
    Base class for file writing operations with support for combined output files.
    
    Class Attributes
    ----------------
    _fp : TextIO | None
        Shared file pointer for combined output mode
    _fp_count : int
        Reference counter for combined file pointer
    _combined_filename : str
        Default filename for combined output
    _combined_filemode : str
        Default file mode for combined output
    """
    _fp = None
    _fp_count = 0
    _logger = logging.getLogger("BaseWriter")

    _combined_filename = "combined.txt"
    _combined_filemode = "w"

    @classmethod
    def basicConfig(cls, combined_filename=None, combined_mode=None):
        """Configure global settings for combined file output."""
        if combined_filename:
            cls._combined_filename = combined_filename
        if combined_mode:
            cls._combined_filemode = combined_mode

    def __init__(self, filename=None, mode="w", combined=False):
        """
        Initialize writer with file handling options.

        Parameters
        ----------
        filename : str | None
            Path to output file
        mode : str
            File open mode ('w' for write, 'a' for append)
        combined : bool
            If True, writes to a shared combined file instead of individual files
        """
        self._logger = logging.getLogger(__class__.__name__)
        self.filename = filename
        self.mode = mode
        self.combined = combined
        self.fp = None
        self.tag = ""  # Prefix for lines in combined mode

        self._logger.debug(f"initialize: file={self.filename}, mode=\"{self.mode}\", combined={self.combined}")
        self._open()

    def __del__(self):
        self._logger.debug(f"finalize: destructor called, file={self.filename}")
        self._close()

    def close(self):
        self._close()

    def _write(self, *args):
        """
        Internal write method handling both combined and individual file modes.
        
        In combined mode, prefixes each line with the source filename tag.
        """
        if self.fp:
            if self.combined:
                # Add filename tag to each line in combined mode
                for arg in args:
                    for s in arg.split("\n"):
                        self.fp.write(self.tag + s + "\n")
            else:
                # Direct write in individual file mode
                self.fp.write("\n".join(args) + "\n")

    def _open(self):
        if self.combined:
            self._logger.debug(f"open: combined file")
            self.fp = __class__._open_combined()
            self.tag = f"<{self.filename}> "
        else:
            self._logger.debug(f"open: local file, file={self.filename}")
            if self.filename is not None:
                self.fp = open(self.filename, self.mode)
            self.tag = ""
        self._logger.debug(f"open: tag=\"{self.tag}\"")

    def _close(self):
        if self.fp:
            if self.combined:
                __class__._close_combined()
            else:
                self.fp.close()
                self._logger.debug(f"close: closed, file={self.filename}")
            self.fp = None

    @classmethod
    def _open_combined(cls):
        if cls._fp is None:
            cls._logger.debug(f"open_combined: open file {cls._combined_filename} with mode=\"{cls._combined_filemode}\"")
            cls._fp = open(cls._combined_filename, cls._combined_filemode)
        else:
            cls._logger.debug(f"open_combined: already opened")
        cls._fp_count += 1
        cls._logger.debug(f"open_combined: increment counter to {cls._fp_count}")
        return cls._fp

    @classmethod
    def _close_combined(cls):
        if cls._fp:
            cls._fp_count -= 1
            cls._logger.debug(f"close_combined: decrement counter to {cls._fp_count}")
            if cls._fp_count <= 0:
                cls._fp.close()
                cls._fp = None
                cls._logger.debug("close_combined: closed")
        

class TextWriter(BaseWriter):
    def __init__(self, filename=None, mode="w", *, combined=False):
        self._logger = logging.getLogger(__class__.__name__)
        self._logger.debug(f"initialize: file={filename}, mode=\"{mode}\", combined={combined}")
        super().__init__(filename=filename, mode=mode, combined=combined)

    def write(self, *args):
        self._write(*args)


class DataWriter(BaseWriter):
    """
    Specialized writer for structured data output with header support.
    
    Extends BaseWriter to handle formatted data output with column headers
    and optional descriptions.
    """
    def __init__(self, filename=None, mode="w", item_list=[], *, 
                 description=None, long_format=False, combined=False):
        """
        Initialize data writer with column specifications.

        Parameters
        ----------
        filename : str | None
            Output file path
        mode : str
            File open mode
        item_list : list
            List of column specifications (name, format, description)
        description : str | None
            Optional file header description
        long_format : bool
            If True, writes detailed column descriptions
        combined : bool
            If True, writes to shared combined file
        """
        self._logger = logging.getLogger(__class__.__name__)
        super().__init__(filename=filename, mode=mode, combined=combined)
        self.header = self._find_item_list(item_list)
        if mode == "w":
            self._write_header(description, long_format)

    def write(self, *args):
        self._write_items(args)

    def _write_header(self, description, long_format):
        """
        Write file header with column descriptions.
        
        Parameters
        ----------
        description : str | None
            Optional file description
        long_format : bool
            If True, writes detailed per-column descriptions
        """
        if description:
            for s in description.split("\n"):
                self._write("# " + s)
        
        if long_format:
            # Write detailed column descriptions with indices
            for idx, (s,_,w) in enumerate(self.header, 1):
                msg = f"# {idx}: " + (w if w else s)
                self._write(msg)
        else:
            # Write simple column names
            self._write("# " + " ".join([s for s,_,_ in self.header]))

    def _write_items(self, items):
        """
        Write a row of data values with proper formatting.
        
        Parameters
        ----------
        items : sequence
            Values to write, must match header length
        """
        assert len(items) == len(self.header)
        msg = []
        for (_, fmt, _), v in zip(self.header, items):
            # Apply format if specified, otherwise convert to string
            msg.append(fmt.format(v) if fmt else str(v))
        self._write(" ".join(msg))
        
    def _find_item_list(self, item_list):
        items = []
        for v in item_list:
            if isinstance(v, str):
                items.append((v, None, None))
            elif (isinstance(v, list) or isinstance(v, tuple)) and len(v) == 3:
                items.append(tuple(v))
            else:
                raise ValueError(f"unknown item {v}")
        return items



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

from typing import List, Union, Optional, TYPE_CHECKING
import time

import numpy as np
import scipy
from scipy.optimize import minimize

import odatse
import odatse.domain

if TYPE_CHECKING:
    from mpi4py import MPI


class Algorithm(odatse.algorithm.AlgorithmBase):
    """
    Algorithm class for performing minimization using the Nelder-Mead method.
    """

    # inputs
    label_list: np.ndarray
    initial_list: np.ndarray
    min_list: np.ndarray
    max_list: np.ndarray
    unit_list: np.ndarray

    # hyperparameters of Nelder-Mead
    initial_simplex_list: List[List[float]]
    xtol: float
    ftol: float

    # results
    xopt: np.ndarray
    fopt: float
    itera: int
    funcalls: int
    allvecs: List[np.ndarray]

    iter_history: List[List[Union[int,float]]]
    fev_history: List[List[Union[int,float]]]

    def __init__(
        self,
        info: odatse.Info,
        runner: odatse.Runner = None,
        domain=None,
        run_mode: str = "initial",
        mpicomm: Optional["MPI.Comm"] = None,
    ) -> None:
        """
        Initialize the Algorithm class.

        Parameters
        ----------
        info : Info
            Information object containing algorithm settings.
        runner : Runner
            Runner object for submitting jobs.
        domain :
            Domain object defining the search space.
        run_mode : str
            Mode of running the algorithm.
        mpicomm : MPI.Comm
            MPI communicator to use for parallelization.
            If not provided, the default MPI communicator (MPI.COMM_WORLD) will be used if mpi4py is installed.
        """
        super().__init__(info=info, runner=runner, run_mode=run_mode, mpicomm=mpicomm)

        if domain and isinstance(domain, odatse.domain.Region):
            self.domain = domain
        else:
            self.domain = odatse.domain.Region(info)

        self.min_list = self.domain.min_list
        self.max_list = self.domain.max_list
        self.unit_list = self.domain.unit_list

        self.domain.initialize(rng=self.rng, limitation=runner.limitation, num_walkers=self.mpisize)
        self.initial_list = self.domain.initial_list[self.mpirank]

        info_minimize = info.algorithm.get("minimize", {})
        self.initial_scale_list = info_minimize.get(
            "initial_scale_list", [0.25] * self.dimension
        )
        self.xtol = info_minimize.get("xatol", 0.0001)
        self.ftol = info_minimize.get("fatol", 0.0001)
        self.maxiter = info_minimize.get("maxiter", 10000)
        self.maxfev = info_minimize.get("maxfev", 100000)

        self._show_parameters()

    def _run(self) -> None:
        """
        Run the minimization algorithm.
        """
        run = self.runner

        min_list = self.min_list
        max_list = self.max_list
        unit_list = self.unit_list
        label_list = self.label_list

        step = [0]
        iter_history = []
        fev_history = []

        f0 = run.submit(self.initial_list, (0, 0))
        iter_history.append([*self.initial_list, f0])

        scipy_version = [int(s) for s in scipy.__version__.split('.')]

        if scipy_version[0] >= 1 and scipy_version[1] >= 11:
            def _cb(intermediate_result):
                """
                Callback function for scipy.optimize.minimize.
                """
                x = intermediate_result.x
                fun = intermediate_result.fun
                print("eval: x={}, fun={}".format(x, fun))
                iter_history.append([*x, fun])
        else:
            def _cb(x):
                """
                Callback function for scipy.optimize.minimize.
                """
                fun = _f_calc(x, 1)
                print("eval: x={}, fun={}".format(x, fun))
                iter_history.append([*x, fun])

        def _f_calc(x_list: np.ndarray, iset) -> float:
            """
            Calculate the objective function value.

            Parameters
            ----------
            x_list : np.ndarray
                List of variables.
            iset :
                Set index.

            Returns
            -------
            float
                Objective function value.
            """
            # check if within region -> boundary option in minimize
            # note: 'bounds' option supported in scipy >= 1.7.0
            in_range = np.all((min_list < x_list) & (x_list < max_list))
            if not in_range:
                print("Warning: out of range: {}".format(x_list))
                return float("inf")

            # check if limitation satisfied
            in_limit = self.runner.limitation.judge(x_list)
            if not in_limit:
                print("Warning: variables do not satisfy the constraint formula")
                return float("inf")

            x_list /= unit_list

            step[0] += 1
            args = (step[0], iset)
            y = run.submit(x_list, args)
            if iset == 0:
                fev_history.append([step[0], *x_list, y])
            return y

        time_sta = time.perf_counter()
        optres = minimize(
            _f_calc,
            self.initial_list,
            method="Nelder-Mead",
            args=(0,),
            # bounds=[(a,b) for a,b in zip(min_list, max_list)],
            options={
                "xatol": self.xtol,
                "fatol": self.ftol,
                "return_all": True,
                "disp": True,
                "maxiter": self.maxiter,
                "maxfev": self.maxfev,
                "initial_simplex": self.initial_simplex_list,
            },
            callback=_cb,
        )

        self.xopt = optres.x
        self.fopt = optres.fun
        self.itera = optres.nit
        self.funcalls = optres.nfev
        self.allvecs = optres.allvecs
        time_end = time.perf_counter()
        self.timer["run"]["min_search"] = time_end - time_sta

        self.iter_history = iter_history
        self.fev_history = fev_history

        self._output_results()

        if self.mpisize > 1:
            self.mpicomm.barrier()

    def _prepare(self):
        """
        Prepare the initial simplex for the Nelder-Mead algorithm.
        """
        # make initial simplex
        #   [ v0, v0+a_1*e_1, v0+a_2*e_2, ... v0+a_d*e_d ]
        # where a = ( a_1 a_2 a_3 ... a_d ) and e_k is a unit vector along k-axis
        v = np.array(self.initial_list)
        a = np.array(self.initial_scale_list)
        self.initial_simplex_list = np.vstack((v, v + np.diag(a)))

    def _output_results(self):
        """
        Output the results of the minimization to files.
        """
        label_list = self.label_list

        with open("SimplexData.txt", "w") as fp:
            fp.write("#step " + " ".join(label_list) + " R-factor\n")
            for i, v in enumerate(self.iter_history):
                fp.write(str(i) + " " + " ".join(map(str,v)) + "\n")

        with open("History_FunctionCall.txt", "w") as fp:
            fp.write("#No " + " ".join(label_list) + "\n")
            for i, v in enumerate(self.fev_history):
                fp.write(" ".join(map(str,v)) + "\n")

        with open("res.txt", "w") as fp:
            fp.write(f"fx = {self.fopt}\n")
            for x, y in zip(label_list, self.xopt):
                fp.write(f"{x} = {y}\n")
            fp.write(f"iterations = {self.itera}\n")
            fp.write(f"function_evaluations = {self.funcalls}\n")

    def _post(self):
        """
        Post-process the results after minimization.
        """
        result = {
            "x": self.xopt,
            "fx": self.fopt,
            "x0": self.initial_list,
        }

        if self.mpisize > 1:
            results = self.mpicomm.allgather(result)
        else:
            results = [result]

        xs = [v["x"] for v in results]
        fxs = [v["fx"] for v in results]
        x0s = [v["x0"] for v in results]

        idx = np.argmin(fxs)

        if self.mpirank == 0:
            label_list = self.label_list
            with open("res.txt", "w") as fp:
                fp.write(f"fx = {fxs[idx]}\n")
                for x, y in zip(label_list, xs[idx]):
                    fp.write(f"{x} = {y}\n")
                if len(results) > 1:
                    fp.write(f"index = {idx}\n")
                    for x, y in zip(label_list, x0s[idx]):
                        fp.write(f"initial {x} = {y}\n")

        return {"x": xs[idx], "fx": fxs[idx], "x0": x0s[idx]}



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

try:
    from mpi4py import MPI
    Comm = MPI.Comm

    __comm = MPI.COMM_WORLD
    __size = __comm.size
    __rank = __comm.rank

    def comm() -> MPI.Comm:
        return __comm

    def size() -> int:
        return __size

    def rank() -> int:
        return __rank

    def enabled() -> bool:
        return True


except ImportError:
    Comm = None

    def comm() -> None:
        return None

    def size() -> int:
        return 1

    def rank() -> int:
        return 0

    def enabled() -> bool:
        return False



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

from abc import ABCMeta, abstractmethod

import numpy as np

from .read_matrix import read_matrix, read_vector


class LimitationBase(metaclass=ABCMeta):
    """
    Abstract base class for limitations.
    """

    @abstractmethod
    def __init__(self, is_limitary: bool):
        """
        Initialize the limitation.

        Parameters
        ----------
        is_limitary : bool
            Boolean indicating if the limitation is active.
        """
        self.is_limitary = is_limitary

    @abstractmethod
    def judge(self, x: np.ndarray) -> bool:
        """
        Abstract method to judge if the limitation is satisfied.

        Parameters
        ----------
        x : np.ndarray
            Input array to be judged.

        Returns
        -------
        bool
            Boolean indicating if the limitation is satisfied.
        """
        raise NotImplementedError

class Unlimited(LimitationBase):
    """
    Class representing an unlimited (no limitation) condition.
    """

    def __init__(self):
        """
        Initialize the unlimited condition.
        """
        super().__init__(False)

    def judge(self, x: np.ndarray) -> bool:
        """
        Always returns True as there is no limitation.

        Parameters
        ----------
        x : np.ndarray
            Input array to be judged.

        Returns
        -------
        bool
            Always True.
        """
        return True

class Inequality(LimitationBase):
    """
    Class representing an inequality limitation.
    """

    def __init__(self, a: np.ndarray, b: np.ndarray, is_limitary: bool):
        """
        Initialize the inequality limitation.

        Parameters
        ----------
        a : np.ndarray
            Coefficient matrix.
        b : np.ndarray
            Constant vector.
        is_limitary : bool
            Boolean indicating if the limitation is active.
        """
        super().__init__(is_limitary)
        if self.is_limitary:
            self.a = np.array(a)
            self.b = np.array(b)
            self.minusb = -np.array(b)
            self.n_formula = a.shape[0]
            self.ndim = a.shape[1]

    def judge(self, x: np.ndarray) -> bool:
        """
        Judge if the inequality limitation is satisfied.

        Parameters
        ----------
        x : np.ndarray
            Input array to be judged.

        Returns
        -------
        bool
            Boolean indicating if the limitation is satisfied.
        """
        if self.is_limitary:
            Ax_b = np.dot(self.a, x) + self.b
            judge_result = np.all(Ax_b > 0)
        else:
            judge_result = True
        return judge_result

    @classmethod
    def from_dict(cls, d):
        """
        Create an Inequality instance from a dictionary.

        Parameters
        ----------
        d
            Dictionary containing 'co_a' and 'co_b' keys.

        Returns
        -------
        Inequality
            an Inequality instance.
        """
        co_a: np.ndarray = read_matrix(d.get("co_a", []))
        co_b: np.ndarray = read_matrix(d.get("co_b", []))

        if co_a.size == 0:
            is_set_co_a = False
        else:
            if co_a.ndim == 2:
                is_set_co_a = True
            else:
                raise ValueError("co_a should be a matrix of size equal to number of constraints times dimension")

        if co_b.size == 0:
            is_set_co_b = False
        else:
            if co_b.ndim == 2 and co_b.shape == (co_a.shape[0], 1):
                is_set_co_b = True
            else:
                raise ValueError("co_b should be a column vector of size equal to number of constraints")

        if is_set_co_a and is_set_co_b:
            is_limitary = True
        elif (not is_set_co_a) and (not is_set_co_b):
            is_limitary = False
        else:
            msg = "ERROR: Both co_a and co_b must be defined."
            raise ValueError(msg)

        return cls(co_a, co_b.reshape(-1), is_limitary)



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

from typing import List, Dict, Union, Any

from pathlib import Path
import numpy as np

import odatse
from ._domain import DomainBase

class MeshGrid(DomainBase):
    """
    MeshGrid class for handling grid data for quantum beam diffraction experiments.
    """

    grid: List[List[Union[int, float]]] = []
    grid_local: List[List[Union[int, float]]] = []
    candicates: int

    def __init__(
        self,
        info: odatse.Info = None,
        *,
        param: Dict[str, Any] = None,
        mesh: bool = True,
        rng: np.random.RandomState = None,
    ):
        """
        Initialize the MeshGrid object.

        Parameters
        ----------
        info : Info, optional
            Information object containing algorithm parameters.
        param : dict, optional
            Dictionary containing parameters for setting up the grid.
        mesh : bool, optional
            Whether to use mesh grid or points.
        rng : np.random.RandomState, optional
            Random number generator.
        """
        super().__init__(info)

        if info:
            if "param" in info.algorithm:
                self._setup(info.algorithm["param"], rng, mesh=mesh)
            else:
                raise ValueError("ERROR: algorithm.param not defined")
        elif param:
            self._setup(param, rng, mesh=mesh)
        else:
            pass

    def do_split(self):
        """
        Split the grid data among MPI processes.
        """
        if self.mpisize > 1:
            index = [idx for idx, *v in self.grid]
            index_local = np.array_split(index, self.mpisize)[self.mpirank]
            self.grid_local = [[idx, *v] for idx, *v in self.grid if idx in index_local]
        else:
            self.grid_local = self.grid

    def _setup(self, info_param, rng: np.random.RandomState, mesh: bool = True):
        """
        Setup the grid based on provided parameters.

        Parameters
        ----------
        info_param
            Dictionary containing parameters for setting up the grid.
        rng : np.random.RandomState, optional
            Random number generator.
        mesh : bool, optional
            Whether to use mesh grid or points.
        """
        if "mesh_path" in info_param:
            self._setup_from_file(info_param)
        elif mesh:
            self._setup_grid(info_param)
        else:
            self._setup_random(info_param, rng)

        self.ncandicates = len(self.grid)

    def _setup_from_file(self, info_param):
        """
        Setup the grid from a file.

        Parameters
        ----------
        info_param
            Dictionary containing parameters for setting up the grid.
        """
        if "mesh_path" not in info_param:
            raise ValueError("ERROR: mesh_path not defined")
        mesh_path = self.root_dir / Path(info_param["mesh_path"]).expanduser()

        if not mesh_path.exists():
            raise FileNotFoundError("mesh_path not found: {}".format(mesh_path))

        comments = info_param.get("comments", "#")
        delimiter = info_param.get("delimiter", None)
        skiprows = info_param.get("skiprows", 0)

        if self.mpirank == 0:
            data = np.loadtxt(mesh_path, comments=comments, delimiter=delimiter, skiprows=skiprows)
            if data.ndim == 1:
                data = data.reshape(1, -1)

            # old format: index x1 x2 ... -> omit index
            data = data[:, 1:]
        else:
            data = None

        if self.mpisize > 1:
            data = odatse.mpi.comm().bcast(data, root=0)

        self.grid = [[idx, *v] for idx, v in enumerate(data)]

    def _setup_grid(self, info_param):
        """
        Setup the grid based on min, max, and num lists.

        Parameters
        ----------
        info_param
            Dictionary containing parameters for setting up the grid.
        """
        if "min_list" not in info_param:
            raise ValueError("ERROR: algorithm.param.min_list is not defined in the input")
        min_list = np.array(info_param["min_list"], dtype=float)

        if "max_list" not in info_param:
            raise ValueError("ERROR: algorithm.param.max_list is not defined in the input")
        max_list = np.array(info_param["max_list"], dtype=float)

        if "num_list" not in info_param:
            raise ValueError("ERROR: algorithm.param.num_list is not defined in the input")
        num_list = np.array(info_param["num_list"], dtype=int)

        if len(min_list) != len(max_list) or len(min_list) != len(num_list):
            raise ValueError("ERROR: lengths of min_list, max_list, num_list do not match")
        xs = [
            np.linspace(mn, mx, num=nm)
            for mn, mx, nm in zip(min_list, max_list, num_list)
        ]
        self.grid = [
            [idx, *v] for idx, v in enumerate(
                np.array(
                    np.meshgrid(*xs, indexing='xy')
                ).reshape(len(xs), -1).transpose()
            )
        ]

    def _setup_random(self, info_param, rng: np.random.RandomState):
        if "min_list" not in info_param:
            raise ValueError("ERROR: algorithm.param.min_list is not defined in the input")
        min_list = np.array(info_param["min_list"], dtype=float)

        if "max_list" not in info_param:
            raise ValueError("ERROR: algorithm.param.max_list is not defined in the input")
        max_list = np.array(info_param["max_list"], dtype=float)

        if "num_points" not in info_param:
            raise ValueError("ERROR: algorithm.param.num_points is not defined in the input")
        num_points = info_param["num_points"]

        if len(min_list) != len(max_list):
            raise ValueError("ERROR: lengths of min_list and max_list do not match")
        if num_points <= 0:
            raise ValueError("ERROR: num_points must be positive")

        local_index = np.array_split(np.arange(num_points), self.mpisize)[self.mpirank]
        num_local = len(local_index)

        self.grid_local = [
            [idx, *v]
            for idx, v in zip(
                local_index,
                rng.uniform(min_list, max_list, size=(num_local, len(min_list))),
            )
        ]

        if self.mpisize > 1:
            grids = odatse.mpi.comm().allgather(self.grid_local)
            self.grid = [v for vs in grids for v in vs]
        else:
            self.grid = self.grid_local

    def store_file(self, store_path, *, header=""):
        """
        Store the grid data to a file.

        Parameters
        ----------
        store_path
            Path to the file where the grid data will be stored.
        header
            Header to be included in the file.
        """
        if self.mpirank == 0:
            np.savetxt(store_path, [[*v] for idx, *v in self.grid], header=header)

    @classmethod
    def from_file(cls, mesh_path):
        """
        Create a MeshGrid object from a file.

        Parameters
        ----------
        mesh_path
            Path to the file containing the grid data.

        Returns
        -------
        MeshGrid
            a MeshGrid object.
        """
        return cls(param={"mesh_path": mesh_path})

    @classmethod
    def from_dict(cls, param):
        """
        Create a MeshGrid object from a dictionary of parameters.

        Parameters
        ----------
        param
            Dictionary containing parameters for setting up the grid.

        Returns
        -------
        MeshGrid
            a MeshGrid object.
        """
        return cls(param=param)


if __name__ == "__main__":
    ms = MeshGrid.from_dict({
        'min_list': [0,0,0],
        'max_list': [1,1,1],
        'num_list': [5,5,5],
    })
    ms.store_file("meshfile.dat", header="sample mesh data")
    
    ms2 = MeshGrid.from_file("meshfile.dat")
    ms2.do_split()

    if odatse.mpi.rank() == 0:
        print(ms2.grid)
    print(odatse.mpi.rank(), ms2.grid_local)

    ms2.store_file("meshfile2.dat", header="store again")



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

from typing import MutableMapping, Optional
from pathlib import Path
from fnmatch import fnmatch

from .util import toml
from . import mpi
from . import exception


class Info:
    """
    A class to represent the information structure for the data-analysis software.
    """

    base: dict
    algorithm: dict
    solver: dict
    runner: dict

    def __init__(self, d: Optional[MutableMapping] = None):
        """
        Initialize the Info object.

        Parameters
        ----------
        d : MutableMapping (optional)
            A dictionary to initialize the Info object.
        """
        if d is not None:
            self.from_dict(d)
        else:
            self._cleanup()

    def from_dict(self, d: MutableMapping) -> None:
        """
        Initialize the Info object from a dictionary.

        Parameters
        ----------
        d : MutableMapping
            A dictionary containing the information to initialize the Info object.

        Raises
        ------
        exception.InputError
            If any required section is missing in the input dictionary.
        """
        for section in ["base", "algorithm", "solver"]:
            if section not in d:
                raise exception.InputError(
                    f"ERROR: section {section} does not appear in input"
                )
        self._cleanup()
        self.base = d["base"]
        self.algorithm = d["algorithm"]
        self.solver = d["solver"]
        self.runner = d.get("runner", {})

        self.base["root_dir"] = (
            Path(self.base.get("root_dir", ".")).expanduser().absolute()
        )
        self.base["output_dir"] = (
            self.base["root_dir"]
            / Path(self.base.get("output_dir", ".")).expanduser()
        )

    def _cleanup(self) -> None:
        """
        Reset the Info object to its default state.
        """
        self.base = {}
        self.base["root_dir"] = Path(".").absolute()
        self.base["output_dir"] = self.base["root_dir"]
        self.algorithm = {}
        self.solver = {}
        self.runner = {}

    @classmethod
    def from_file(cls, file_name, fmt="", **kwargs):
        """
        Create an Info object from a file.

        Parameters
        ----------
        file_name : str
            The name of the file to load the information from.
        fmt : str
            The format of the file (default is "").
        **kwargs
            Additional keyword arguments.

        Returns
        -------
        Info
            An Info object initialized with the data from the file.

        Raises
        ------
        ValueError
            If the file format is unsupported.
        """
        if fmt == "toml" or fnmatch(file_name.lower(), "*.toml"):
            inp = {}
            if mpi.rank() == 0:
                inp = toml.load(file_name)
            if mpi.size() > 1:
                inp = mpi.comm().bcast(inp, root=0)
            return cls(inp)
        else:
            raise ValueError("unsupported file format: {}".format(file_name))



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import odatse

def initialize():
    """
    Initialize for main function by parsing commandline arguments and loading input files

    Returns
    -------
    Tuple(Info, str)
        an Info object having parameter values, and a run_mode string
    """
    import argparse

    parser = argparse.ArgumentParser(
        description=(
            "Data-analysis software of quantum beam "
            "diffraction experiments for 2D material structure"
        )
    )
    parser.add_argument("inputfile", help="input file with TOML format")
    parser.add_argument("--version", action="version", version=odatse.__version__)

    mode_group = parser.add_mutually_exclusive_group()
    mode_group.add_argument("--init", action="store_true", help="initial start (default)")
    mode_group.add_argument("--resume", action="store_true", help="resume intterupted run")
    mode_group.add_argument("--cont", action="store_true", help="continue from previous run")

    parser.add_argument("--reset_rand", action="store_true", default=False, help="new random number series in resume or continue mode")

    args = parser.parse_args()


    if args.init is True:
        run_mode = "initial"
    elif args.resume is True:
        run_mode = "resume"
        if args.reset_rand is True:
            run_mode = "resume-resetrand"
    elif args.cont is True:
        run_mode = "continue"
        if args.reset_rand is True:
            run_mode = "continue-resetrand"
    else:
        run_mode = "initial"  # default

    info = odatse.Info.from_file(args.inputfile)
    # info.algorithm.update({"run_mode": run_mode})

    return info, run_mode



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import typing
from typing import Union, List, Iterable

import abc

import collections
import itertools
import numpy as np


class Resampler(abc.ABC):
    @abc.abstractmethod
    def reset(self, weights: Iterable):
        ...

    @abc.abstractmethod
    def sample(self, rs: np.random.RandomState, size=None) -> Union[int, np.ndarray]:
        ...


class BinarySearch(Resampler):
    """
    A resampler that uses binary search to sample based on given weights.
    """
    weights_accumulated: List[float]
    wmax: float

    def __init__(self, weights: Iterable):
        """
        Initialize the BinarySearch resampler with the given weights.

        Parameters
        ----------
        weights : Iterable
            An iterable of weights.
        """
        self.reset(weights)

    def reset(self, weights: Iterable):
        """
        Reset the resampler with new weights.

        Parameters
        ----------
        weights : Iterable
            An iterable of weights.
        """
        self.weights_accumulated = list(itertools.accumulate(weights))
        self.wmax = self.weights_accumulated[-1]

    @typing.overload
    def sample(self, rs: np.random.RandomState) -> int:
        """
        Sample a single index based on the weights.

        Parameters
        ----------
        rs : np.random.RandomState
            A random state for generating random numbers.

        Returns
        -------
        int
            A single sampled index.
        """
        ...

    @typing.overload
    def sample(self, rs: np.random.RandomState, size) -> np.ndarray:
        """
        Sample multiple indices based on the weights.

        Parameters
        ----------
        rs : np.random.RandomState
            A random state for generating random numbers.
        size :
            The number of samples to generate.

        Returns
        -------
        np.ndarray
            An array of sampled indices.
        """
        ...

    def sample(self, rs: np.random.RandomState, size=None) -> Union[int, np.ndarray]:
        """
        Sample indices based on the weights.

        Parameters
        ----------
        rs : np.random.RandomState
            A random state for generating random numbers.
        size :
            The number of samples to generate. If None, a single sample is generated.

        Returns
        -------
        int or np.ndarray
            A single sampled index or an array of sampled indices.
        """
        if size is None:
            return self._sample(self.wmax * rs.rand())
        else:
            return np.array([self._sample(r) for r in self.wmax * rs.rand(size)])

    def _sample(self, r: float) -> int:
        """
        Perform a binary search to find the index corresponding to the given random number.

        Parameters
        ----------
        r : float
            A random number scaled by the maximum weight.

        Returns
        -------
        int
            The index corresponding to the random number.
        """
        return typing.cast(int, np.searchsorted(self.weights_accumulated, r))


class WalkerTable(Resampler):
    """
    A resampler that uses Walker's alias method to sample based on given weights.
    """
    N: int
    itable: np.ndarray
    ptable: np.ndarray

    def __init__(self, weights: Iterable):
        """
        Initialize the WalkerTable resampler with the given weights.

        Parameters
        ----------
        weights : Iterable
            An iterable of weights.
        """
        self.reset(weights)

    def reset(self, weights: Iterable):
        """
        Reset the resampler with new weights.

        Parameters
        ----------
        weights : Iterable
            An iterable of weights.
        """
        self.ptable = np.array(weights).astype(np.float64).flatten()
        self.N = len(self.ptable)
        self.itable = np.full(self.N, -1)
        mean = self.ptable.mean()
        self.ptable -= mean
        shorter = collections.deque([i for i, p in enumerate(self.ptable) if p < 0.0])
        longer = collections.deque([i for i, p in enumerate(self.ptable) if p >= 0.0])

        while len(longer) > 0 and len(shorter) > 0:
            ilong = longer[0]
            ishort = shorter.popleft()
            self.itable[ishort] = ilong
            self.ptable[ilong] += self.ptable[ishort]
            if self.ptable[ilong] <= 0.0:
                longer.popleft()
                shorter.append(ilong)
        self.ptable += mean
        self.ptable *= 1.0 / mean

    @typing.overload
    def sample(self, rs: np.random.RandomState) -> int:
        """
        Sample a single index based on the weights.

        Parameters
        ----------
        rs : np.random.RandomState
            A random state for generating random numbers.

        Returns
        -------
        int
            A single sampled index.
        """
        ...

    @typing.overload
    def sample(self, rs: np.random.RandomState, size) -> np.ndarray:
        """
        Sample multiple indices based on the weights.

        Parameters
        ----------
        rs : np.random.RandomState
            A random state for generating random numbers.
        size :
            The number of samples to generate.

        Returns
        -------
        np.ndarray
            An array of sampled indices.
        """
        ...

    def sample(self, rs: np.random.RandomState, size=None) -> Union[int, np.ndarray]:
        """
        Sample indices based on the weights.

        Parameters
        ----------
        rs : np.random.RandomState
            A random state for generating random numbers.
        size :
            The number of samples to generate. If None, a single sample is generated.

        Returns
        -------
        int or np.ndarray
            A single sampled index or an array of sampled indices.
        """
        if size is None:
            r = rs.rand() * self.N
            return self._sample(r)
        else:
            r = rs.rand(size) * self.N
            i = np.floor(r).astype(np.int64)
            p = r - i
            ret = np.where(p < self.ptable[i], i, self.itable[i])
            return ret

    def _sample(self, r: float) -> int:
        """
        Perform a sampling operation based on the given random number.

        Parameters
        ----------
        r : float
            A random number scaled by the number of weights.

        Returns
        -------
        int
            The index corresponding to the random number.
        """
        i = int(np.floor(r))
        p = r - i
        if p < self.ptable[i]:
            return i
        else:
            return self.itable[i]

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(prog="resampling test")
    parser.add_argument(
        "-s", "--seed", type=int, default=12345, help="random number seed"
    )
    parser.add_argument("-m", "--method", default="walker", help="method to resample")
    parser.add_argument("-N", type=int, default=100000, help="Number of samples")

    args = parser.parse_args()

    rs = np.random.RandomState(args.seed)

    ps = [0.0, 1.0, 2.0, 3.0]
    # ps = rs.rand(5)
    S = np.sum(ps)
    if args.method == "walker":
        resampler = WalkerTable(ps)
    else:
        resampler = BinarySearch(ps)
    samples = resampler.sample(rs, args.N)

    print("#i result exact diff")
    for i, p in enumerate(ps):
        r = np.count_nonzero(samples == i) / args.N
        print(f"{i} {r} {p/S} {r - p/S}")



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

from ._algorithm import AlgorithmBase

def choose_algorithm(name):
    """
    Search for algorithm module by name

    Parameters
    ----------
    name : str
        name of the algorithm

    Returns
    -------
    module
        algorithm module
    """

    alg_table = {
        "mapper": "mapper_mpi",
        "minsearch": "min_search",
    }

    try:
        import importlib
        alg_name = "odatse.algorithm.{}".format(alg_table.get(name, name))
        alg_module = importlib.import_module(alg_name)
    except ModuleNotFoundError as e:
        print("ERROR: {}".format(e))
        import sys
        sys.exit(1)

    return alg_module



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.


from abc import ABCMeta, abstractmethod
from enum import IntEnum
import time
import os
import pathlib
import pickle
import shutil
import copy

import numpy as np

import odatse
import odatse.util.limitation
from odatse import exception, mpi

# for type hints
from pathlib import Path
from typing import List, Optional, TYPE_CHECKING, Dict, Tuple


if TYPE_CHECKING:
    from mpi4py import MPI

class AlgorithmStatus(IntEnum):
    """Enumeration for the status of the algorithm."""
    INIT = 1
    PREPARE = 2
    RUN = 3

class AlgorithmBase(metaclass=ABCMeta):
    """Base class for algorithms, providing common functionality and structure."""

    mpicomm: Optional["MPI.Comm"]
    mpisize: int
    mpirank: int
    rng: np.random.RandomState
    dimension: int
    label_list: List[str]
    runner: Optional[odatse.Runner]

    root_dir: Path
    output_dir: Path
    proc_dir: Path

    timer: Dict[str, Dict]

    status: AlgorithmStatus = AlgorithmStatus.INIT
    mode: Optional[str] = None

    @abstractmethod
    def __init__(
            self,
            info: odatse.Info,
            runner: Optional[odatse.Runner] = None,
            run_mode: str = "initial",
            mpicomm: Optional["MPI.Comm"] = None
    ) -> None:
        """
        Initialize the algorithm with the given information and runner.

        Parameters
        ----------
        info : Info
            Information object containing algorithm and base parameters.
        runner : Runner (optional)
            Optional runner object to execute the algorithm.
        run_mode : str
            Mode in which the algorithm should run.
        mpicomm : MPI.Comm (optional)
            MPI communicator to use for parallelization.
            If not provided, the default MPI communicator (MPI.COMM_WORLD) will be used if mpi4py is installed.
        """
        if mpicomm is None:
            self.mpicomm = mpi.comm()
            self.mpisize = mpi.size()
            self.mpirank = mpi.rank()
        else:
            self.mpicomm = mpicomm
            self.mpisize = mpicomm.size
            self.mpirank = mpicomm.rank
        self.timer = {"init": {}, "prepare": {}, "run": {}, "post": {}}
        self.timer["init"]["total"] = 0.0
        self.status = AlgorithmStatus.INIT
        self.mode = run_mode.lower()

        # keep copy of parameters
        self.info = copy.deepcopy(info.algorithm)

        self.dimension = info.algorithm.get("dimension") or info.base.get("dimension")
        if not self.dimension:
            raise ValueError("ERROR: dimension is not defined")

        if "label_list" in info.algorithm:
            label = info.algorithm["label_list"]
            if len(label) != self.dimension:
                raise ValueError(f"ERROR: length of label_list and dimension do not match ({len(label)} != {self.dimension})")
            self.label_list = label
        else:
            self.label_list = [f"x{d+1}" for d in range(self.dimension)]

        # initialize random number generator
        self.__init_rng(info)

        # directories
        self.root_dir = info.base["root_dir"]
        self.output_dir = info.base["output_dir"]
        self.proc_dir = self.output_dir / str(self.mpirank)
        self.proc_dir.mkdir(parents=True, exist_ok=True)
        # Some cache of the filesystem may delay making a dictionary
        # especially when mkdir just after removing the old one
        while not self.proc_dir.is_dir():
            time.sleep(0.1)
        if self.mpisize > 1:
            self.mpicomm.Barrier()

        # checkpointing
        self.checkpoint = info.algorithm.get("checkpoint", False)
        self.checkpoint_file = info.algorithm.get("checkpoint_file", "status.pickle")
        self.checkpoint_steps = info.algorithm.get("checkpoint_steps", 65536*256)  # large number
        self.checkpoint_interval = info.algorithm.get("checkpoint_interval", 86400*360)  # longer enough

        # runner
        if runner is not None:
            self.set_runner(runner)

    def __init_rng(self, info: odatse.Info) -> None:
        """
        Initialize the random number generator.

        Parameters
        ----------
        info : Info
            Information object containing algorithm parameters.
        """
        seed = info.algorithm.get("seed", None)
        seed_delta = info.algorithm.get("seed_delta", 314159)

        if seed is None:
            self.rng = np.random.RandomState()
        else:
            self.rng = np.random.RandomState(seed + self.mpirank * seed_delta)

    def set_runner(self, runner: odatse.Runner) -> None:
        """
        Set the runner for the algorithm.

        Parameters
        ----------
        runner : Runner
            Runner object to execute the algorithm.
        """
        self.runner = runner

    def prepare(self) -> None:
        """
        Prepare the algorithm for execution.
        """
        if self.runner is None:
            msg = "Runner is not assigned"
            raise RuntimeError(msg)
        self._prepare()
        self.status = AlgorithmStatus.PREPARE

    @abstractmethod
    def _prepare(self) -> None:
        """Abstract method to be implemented by subclasses for preparation steps."""
        pass

    def run(self) -> None:
        """
        Run the algorithm.
        """
        if self.status < AlgorithmStatus.PREPARE:
            msg = "algorithm has not prepared yet"
            raise RuntimeError(msg)
        original_dir = os.getcwd()
        os.chdir(self.proc_dir)
        self.runner.prepare(self.proc_dir)
        self._run()
        self.runner.post()
        os.chdir(original_dir)
        self.status = AlgorithmStatus.RUN

    @abstractmethod
    def _run(self) -> None:
        """Abstract method to be implemented by subclasses for running steps."""
        pass

    def post(self) -> Dict:
        """
        Perform post-processing after the algorithm has run.

        Returns
        -------
        Dict
            Dictionary containing post-processing results.
        """
        if self.status < AlgorithmStatus.RUN:
            msg = "algorithm has not run yet"
            raise RuntimeError(msg)
        original_dir = os.getcwd()
        os.chdir(self.output_dir)
        result = self._post()
        os.chdir(original_dir)
        return result

    @abstractmethod
    def _post(self) -> Dict:
        """Abstract method to be implemented by subclasses for post-processing steps."""
        pass

    def main(self):
        """
        Main method to execute the algorithm.
        """
        time_sta = time.perf_counter()
        self.prepare()
        time_end = time.perf_counter()
        self.timer["prepare"]["total"] = time_end - time_sta
        if self.mpisize > 1:
            self.mpicomm.Barrier()

        time_sta = time.perf_counter()
        self.run()
        time_end = time.perf_counter()
        self.timer["run"]["total"] = time_end - time_sta
        print("end of run")
        if self.mpisize > 1:
            self.mpicomm.Barrier()

        time_sta = time.perf_counter()
        result = self.post()
        time_end = time.perf_counter()
        self.timer["post"]["total"] = time_end - time_sta

        self.write_timer(self.proc_dir / "time.log")

        return result

    def write_timer(self, filename: Path):
        """
        Write the timing information to a file.

        Parameters
        ----------
        filename : Path
            Path to the file where timing information will be written.
        """
        with open(filename, "w") as fw:
            fw.write("#in units of seconds\n")

            def output_file(type):
                d = self.timer[type]
                fw.write("#{}\n total = {}\n".format(type, d["total"]))
                for key, t in d.items():
                    if key == "total":
                        continue
                    fw.write(" - {} = {}\n".format(key, t))

            output_file("init")
            output_file("prepare")
            output_file("run")
            output_file("post")

    def _save_data(self, data, filename="state.pickle", ngen=3) -> None:
        """
        Save data to a file with versioning.

        Parameters
        ----------
        data
            Data to be saved.
        filename
            Name of the file to save the data.
        ngen : int, default: 3
            Number of generations for versioning.
        """
        try:
            fn = Path(filename + ".tmp")
            with open(fn, "wb") as f:
                pickle.dump(data, f)
        except Exception as e:
            print("ERROR: {}".format(e))
            sys.exit(1)

        for idx in range(ngen-1, 0, -1):
            fn_from = Path(filename + "." + str(idx))
            fn_to = Path(filename + "." + str(idx+1))
            if fn_from.exists():
                shutil.move(fn_from, fn_to)
        if ngen > 0:
            if Path(filename).exists():
                fn_to = Path(filename + "." + str(1))
                shutil.move(Path(filename), fn_to)
        shutil.move(Path(filename + ".tmp"), Path(filename))
        print("save_state: write to {}".format(filename))

    def _load_data(self, filename="state.pickle") -> Dict:
        """
        Load data from a file.

        Parameters
        ----------
        filename
            Name of the file to load the data from.

        Returns
        -------
        Dict
            Dictionary containing the loaded data.
        """
        if Path(filename).exists():
            try:
                fn = Path(filename)
                with open(fn, "rb") as f:
                    data = pickle.load(f)
            except Exception as e:
                print("ERROR: {}".format(e))
                sys.exit(1)
            print("load_state: load from {}".format(filename))
        else:
            print("ERROR: file {} not exist.".format(filename))
            data = {}
        return data

    def _show_parameters(self):
        """
        Show the parameters of the algorithm.
        """
        if self.mpirank == 0:
            info = flatten_dict(self.info)
            for k, v in info.items():
                print("{:16s}: {}".format(k, v))

    def _check_parameters(self, param=None):
        """
        Check the parameters of the algorithm against previous parameters.

        Parameters
        ----------
        param (optional)
            Previous parameters to check against.
        """
        info = flatten_dict(self.info)
        info_prev = flatten_dict(param)

        for k,v in info.items():
            w = info_prev.get(k, None)
            if v != w:
                if self.mpirank == 0:
                    print("WARNING: parameter {} changed from {} to {}".format(k, w, v))
            if self.mpirank == 0:
                print("{:16s}: {}".format(k, v))

# utility
def flatten_dict(d, parent_key="", separator="."):
    """
    Flatten a nested dictionary.

    Parameters
    ----------
    d
        Dictionary to flatten.
    parent_key : str, default : ""
        Key for the parent dictionary.
    separator : str, default : "."
        Separator to use between keys.

    Returns
    -------
    dict
        Flattened dictionary.
    """
    items = []
    if d:
        for key_, val in d.items():
            key = parent_key + separator + key_ if parent_key else key_
            if isinstance(val, dict):
                items.extend(flatten_dict(val, key, separator=separator).items())
            else:
                items.append((key, val))
    return dict(items)



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

from typing import MutableMapping, Any

from ..mpi import rank

try:
    import tomllib
    # print("tomllib is available")
except ImportError:
    tomllib = None
    try:
        import tomli
        if tomli.__version__ < "1.2.0":
            use_old_tomli_api = True
            # print("tomli old api is available")
        else:
            use_old_tomli_api = False
            # print("tomli new api is available")
    except ImportError:
        tomli = None
        try:
            import toml
            # print("toml is available")
            if rank() == 0:
                print("WARNING: use of toml package is left for compatibility.")
                print("         please use tomli package instead.")
                print("HINT: python3 -m pip install tomli")
                print()
        except ImportError:
            print("ERROR: toml parser library is not available.")
            raise


def load(path: str) -> MutableMapping[str, Any]:
    """read TOML file

    Parameters
    ----------
    path: str
        File path to an input TOML file

    Returns
    -------
    toml_dict: MutableMapping[str, Any]
        Dictionary representing TOML file

    """
    if tomllib:
        with open(path, "rb") as f:
            return tomllib.load(f)
    elif tomli:
        if use_old_tomli_api:
            with open(path, "r") as f:
                return tomli.load(f)
        else:
            with open(path, "rb") as f:
                return tomli.load(f)
    elif toml:
        return toml.load(path)



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.


from typing import List

import collections
import numpy as np


def is_connected(nnlist: List[List[int]]) -> bool:
    """
    Check if the graph represented by the neighbor list is connected.

    Parameters
    ----------
    nnlist : List[List[int]]
        A list of lists where each sublist represents the neighbors of a node.

    Returns
    -------
    bool
        True if the graph is connected, False otherwise.
    """
    nnodes = len(nnlist)
    visited = np.full(nnodes, False)
    nvisited = 1
    visited[0] = True
    stack = collections.deque([0])
    while len(stack) > 0:
        node = stack.pop()
        neighbors = [n for n in nnlist[node] if not visited[n]]
        visited[neighbors] = True
        stack.extend(neighbors)
        nvisited += len(neighbors)

    return nvisited == nnodes


def is_bidirectional(nnlist: List[List[int]]) -> bool:
    """
    Check if the graph represented by the neighbor list is bidirectional.

    Parameters
    ----------
    nnlist : List[List[int]]
        A list of lists where each sublist represents the neighbors of a node.

    Returns
    -------
    bool
        True if the graph is bidirectional, False otherwise.
    """
    for i in range(len(nnlist)):
        for j in nnlist[i]:
            if i not in nnlist[j]:
                return False
    return True

if __name__ == "__main__":
    filename = "./neighborlist.txt"
    nnlist = []
    with open(filename) as f:
        for line in f:
            words = line.split()
            nn = [int(w) for w in words[1:]]
            nnlist.append(nn)
    print(is_connected(nnlist))



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import copy
import numpy as np

from .read_matrix import read_matrix

# type hints
from typing import Optional

class MappingBase:
    """
    Base class for mapping operations.
    """

    def __init__(self):
        pass

    def __call__(self, x: np.ndarray) -> np.ndarray:
        """
        Apply the mapping to the input array.

        Parameters
        ----------
        x : np.ndarray
            Input array.

        Returns
        -------
        np.ndarray
            Mapped array.
        """
        raise NotImplementedError


class TrivialMapping(MappingBase):
    """
    A trivial mapping that returns the input array unchanged.
    """

    def __init__(self):
        super().__init__()

    def __call__(self, x: np.ndarray) -> np.ndarray:
        """
        Return the input array unchanged.

        Parameters
        ----------
        x : np.ndarray
            Input array.

        Returns
        -------
        np.ndarray
            The same input array.
        """
        return x


class Affine(MappingBase):
    """
    An affine mapping defined by a matrix A and a vector b.
    """

    A: Optional[np.ndarray]
    b: Optional[np.ndarray]

    def __init__(self, A: Optional[np.ndarray] = None, b: Optional[np.ndarray] = None):
        """
        Initialize the affine mapping.

        Parameters
        ----------
        A : np.ndarray, optional
            Transformation matrix.
        b : np.ndarray, optional
            Translation vector.
        """
        # copy arguments
        self.A = np.array(A) if A is not None else None
        self.b = np.array(b) if b is not None else None

        # check
        if self.A is not None:
            if not self.A.ndim == 2:
                raise ValueError("A is not a matrix")
        if self.b is not None:
            if not self.b.ndim == 1:
                raise ValueError("b is not a vector")
        if self.A is not None and self.b is not None:
            if not self.A.shape[0] == self.b.shape[0]:
                raise ValueError("shape of A and b mismatch")

    def __call__(self, x: np.ndarray) -> np.ndarray:
        """
        Apply the affine mapping to the input array.

        Parameters
        ----------
        x : np.ndarray
            Input array.

        Returns
        -------
        np.ndarray
            Mapped array.
        """
        if self.A is None:
            ret = copy.copy(x)
        else:
            ret = np.dot(self.A, x)
        if self.b is None:
            return ret
        else:
            return ret + self.b

    @classmethod
    def from_dict(cls, d):
        """
        Create an Affine instance from a dictionary.

        Parameters
        ----------
        d : dict
            Dictionary containing 'A' and 'b' keys.

        Returns
        -------
        Affine
            An instance of the Affine class.
        """
        A: Optional[np.ndarray] = read_matrix(d.get("A", []))
        b: Optional[np.ndarray] = read_matrix(d.get("b", []))

        if A is None:
            pass
        elif A.size == 0:
            A = None
        else:
            if not A.ndim == 2:
                raise ValueError("A should be a matrix")

        if b is None:
            pass
        elif b.size == 0:
            b = None
        else:
            if not (b.ndim == 2 and b.shape[1] == 1):
                raise ValueError("b should be a column vector")
            if not (A is not None and b.shape[0] == A.shape[0]):
                raise ValueError("shape of A and b mismatch")
            b = b.reshape(-1)

        return cls(A, b)



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import typing
from typing import List, Set
from os import PathLike

import sys
import itertools

import numpy as np

from odatse import mpi

try:
    from tqdm import tqdm

    has_tqdm = True
except:
    has_tqdm = False


class Cells:
    """
    A class to represent a grid of cells for spatial partitioning.
    
    This class divides a spatial region into a grid of cells to efficiently
    find neighboring points within a specified radius. Each cell contains
    a set of point indices that fall within its spatial boundaries.
    
    Attributes
    ----------
    cells : List[Set[int]]
        List of sets, where each set contains indices of points in that cell.
    dimension : int
        Number of spatial dimensions.
    mins : np.ndarray
        Minimum coordinates of the grid in each dimension.
    maxs : np.ndarray
        Maximum coordinates of the grid in each dimension.
    Ns : np.ndarray
        Number of cells in each dimension.
    ncell : int
        Total number of cells in the grid.
    cellsize : float
        Size of each cell.
    
    Examples
    --------
    >>> mins = np.array([0.0, 0.0, 0.0])
    >>> maxs = np.array([10.0, 10.0, 10.0])
    >>> cells = Cells(mins, maxs, cellsize=2.0)
    >>> point_index = 0
    >>> point_coords = np.array([1.5, 3.2, 4.7])
    >>> cell_index = cells.coord2cellindex(point_coords)
    >>> cells.cells[cell_index].add(point_index)
    """

    cells: List[Set[int]]
    dimension: int
    mins: np.ndarray
    maxs: np.ndarray
    Ns: np.ndarray
    ncell: int
    cellsize: float

    def __init__(self, mins: np.ndarray, maxs: np.ndarray, cellsize: float) -> None:
        """
        Initialize the Cells object.
        
        Parameters
        ----------
        mins : np.ndarray
            The minimum coordinates of the grid in each dimension.
        maxs : np.ndarray
            The maximum coordinates of the grid in each dimension.
        cellsize : float
            The size of each cell.
            
        Returns
        -------
        None
            This method initializes the object's attributes.
            
        Examples
        --------
        >>> mins = np.array([0.0, 0.0, 0.0])
        >>> maxs = np.array([10.0, 10.0, 10.0])
        >>> cells = Cells(mins, maxs, cellsize=2.0)
        """
        self.dimension = len(mins)
        self.mins = mins
        Ls = (maxs - mins) * 1.001
        self.Ns = np.ceil(Ls / cellsize).astype(np.int64)
        self.maxs = self.mins + cellsize * self.Ns
        self.cellsize = cellsize
        self.ncell = typing.cast(int, np.prod(self.Ns))
        self.cells = [set() for _ in range(self.ncell)]

    def coord2cellindex(self, x: np.ndarray) -> int:
        """
        Convert spatial coordinates to a cell index.
        
        This method transforms a point's spatial coordinates into the index
        of the cell containing it, by first converting to cell coordinates
        and then to a cell index.
        
        Parameters
        ----------
        x : np.ndarray
            The spatial coordinates of a point.
            
        Returns
        -------
        int
            The index of the cell containing the point.
            
        Examples
        --------
        >>> cells = Cells(np.array([0, 0]), np.array([10, 10]), 2.0)
        >>> cells.coord2cellindex(np.array([3.5, 4.2]))
        12
        """
        return self.cellcoord2cellindex(self.coord2cellcoord(x))

    def coord2cellcoord(self, x: np.ndarray) -> np.ndarray:
        """
        Convert spatial coordinates to cell coordinates.
        
        Cell coordinates are integer indices that identify the position of a cell
        within the grid along each dimension.
        
        Parameters
        ----------
        x : np.ndarray
            The spatial coordinates of a point.
            
        Returns
        -------
        np.ndarray
            The cell coordinates (integer indices for each dimension).
            
        Examples
        --------
        >>> cells = Cells(np.array([0, 0]), np.array([10, 10]), 2.0)
        >>> cells.coord2cellcoord(np.array([3.5, 4.2]))
        array([1, 2])
        """
        return np.floor((x - self.mins) / self.cellsize).astype(np.int64)

    def cellcoord2cellindex(self, ns: np.ndarray) -> int:
        """
        Convert cell coordinates to a cell index.
        
        This method converts multi-dimensional cell coordinates to a single
        index that uniquely identifies the cell in the flat cells list.
        
        Parameters
        ----------
        ns : np.ndarray
            The cell coordinates (integer indices for each dimension).
            
        Returns
        -------
        int
            The index of the cell in the flat cells list.
            
        Examples
        --------
        >>> cells = Cells(np.array([0, 0]), np.array([10, 10]), 2.0)
        >>> cells.cellcoord2cellindex(np.array([1, 2]))
        12
        """
        index = 0
        oldN = 1
        for n, N in zip(ns, self.Ns):
            index *= oldN
            index += n
            oldN = N
        return index

    def cellindex2cellcoord(self, index: int) -> np.ndarray:
        """
        Convert a cell index to cell coordinates.
        
        This method is the inverse of cellcoord2cellindex and converts a flat cell
        index back to multi-dimensional cell coordinates.
        
        Parameters
        ----------
        index : int
            The index of the cell in the flat cells list.
            
        Returns
        -------
        np.ndarray
            The cell coordinates (integer indices for each dimension).
            
        Examples
        --------
        >>> cells = Cells(np.array([0, 0]), np.array([10, 10]), 2.0)
        >>> cells.cellindex2cellcoord(12)
        array([1, 2])
        """
        ns = np.zeros(self.dimension, dtype=np.int64)
        for d in range(self.dimension):
            d = self.dimension - d - 1
            N = self.Ns[d]
            ns[d] = index % N
            index = index // N
        return ns

    def out_of_bound(self, ns: np.ndarray) -> bool:
        """
        Check if cell coordinates are out of bounds.
        
        This method verifies whether the given cell coordinates are within
        the valid range of the grid.
        
        Parameters
        ----------
        ns : np.ndarray
            The cell coordinates to check.
            
        Returns
        -------
        bool
            True if the coordinates are out of bounds, False otherwise.
            
        Examples
        --------
        >>> cells = Cells(np.array([0, 0]), np.array([10, 10]), 2.0)
        >>> cells.out_of_bound(np.array([-1, 2]))
        True
        >>> cells.out_of_bound(np.array([3, 4]))
        False
        """
        if np.any(ns < 0):
            return True
        if np.any(ns >= self.Ns):
            return True
        return False

    def neighborcells(self, index: int) -> List[int]:
        """
        Get the indices of neighboring cells, including the cell itself.
        
        This method returns the indices of all cells that are adjacent to the specified
        cell in all dimensions, along with the cell itself.
        
        Parameters
        ----------
        index : int
            The index of the cell.
            
        Returns
        -------
        List[int]
            The indices of the neighboring cells (including the cell itself).
            
        Examples
        --------
        >>> cells = Cells(np.array([0, 0]), np.array([10, 10]), 2.0)
        >>> neighbors = cells.neighborcells(12)
        >>> len(neighbors)  # 3x3 neighborhood in 2D
        9
        """
        neighbors: List[int] = []
        center_coord = self.cellindex2cellcoord(index)
        for diff in itertools.product([-1, 0, 1], repeat=self.dimension):
            other_coord = center_coord + np.array(diff)
            if self.out_of_bound(other_coord):
                continue
            other_coord_index = self.cellcoord2cellindex(other_coord)
            neighbors.append(other_coord_index)
        return neighbors

def make_neighbor_list_cell(
    X: np.ndarray,
    radius: float,
    allow_selfloop: bool,
    show_progress: bool,
    comm: mpi.Comm = None,
) -> List[List[int]]:
    """
    Create a neighbor list using cell-based spatial partitioning.
    
    This function uses the Cells class to efficiently find neighboring points
    within the specified radius by only considering points in adjacent cells.
    
    Parameters
    ----------
    X : np.ndarray
        The coordinates of the points (N x D array where N is the number of points
        and D is the dimensionality).
    radius : float
        The radius within which points are considered neighbors.
    allow_selfloop : bool
        Whether to allow a point to be its own neighbor.
    show_progress : bool
        Whether to show a progress bar during computation.
    comm : mpi.Comm, optional
        The MPI communicator for parallel processing, by default None.
        
    Returns
    -------
    List[List[int]]
        A list of lists, where each inner list contains the indices of
        neighboring points for a given point.
        
    Examples
    --------
    >>> coords = np.array([[0.0, 0.0], [1.0, 0.0], [0.0, 1.0], [5.0, 5.0]])
    >>> neighbor_list = make_neighbor_list_cell(coords, radius=1.5, 
    ...                                        allow_selfloop=False, 
    ...                                        show_progress=False)
    >>> neighbor_list
    [[1, 2], [0, 2], [0, 1], []]
    """
    if comm is None:
        mpisize = 1
        mpirank = 0
    else:
        mpisize = comm.size
        mpirank = comm.rank

    mins = typing.cast(np.ndarray, X.min(axis=0))
    maxs = typing.cast(np.ndarray, X.max(axis=0))
    cells = Cells(mins, maxs, radius * 1.001)
    npoints = X.shape[0]
    for n in range(npoints):
        xs = X[n, :]
        index = cells.coord2cellindex(xs)
        cells.cells[index].add(n)

    points = np.array_split(range(npoints), mpisize)[mpirank]
    npoints_local = len(points)
    nnlist: List[List[int]] = [[] for _ in range(npoints_local)]

    if mpirank == 0 and show_progress and has_tqdm:
        desc = "rank 0" if mpisize > 1 else None
        ns = tqdm(points, desc=desc)
    else:
        ns = points

    for n in ns:
        xs = X[n, :]
        cellindex = cells.coord2cellindex(xs)
        for neighborcell in cells.neighborcells(cellindex):
            for other in cells.cells[neighborcell]:
                if (not allow_selfloop) and n == other:
                    continue
                ys = X[other, :]
                r = np.linalg.norm(xs - ys)
                if r <= radius:
                    nnlist[n - points[0]].append(other)
    if mpisize > 1:
        nnlist = list(itertools.chain.from_iterable(comm.allgather(nnlist)))

    nnlist = [sorted(nn) for nn in nnlist]
    return nnlist


def make_neighbor_list_naive(
    X: np.ndarray,
    radius: float,
    allow_selfloop: bool,
    show_progress: bool,
    comm: mpi.Comm = None,
) -> List[List[int]]:
    """
    Create a neighbor list using a naive all-pairs approach.
    
    This function computes the distance between all pairs of points to find
    neighbors within the specified radius. This is less efficient than the
    cell-based approach for large point sets but may be more straightforward.
    
    Parameters
    ----------
    X : np.ndarray
        The coordinates of the points (N x D array where N is the number of points
        and D is the dimensionality).
    radius : float
        The radius within which points are considered neighbors.
    allow_selfloop : bool
        Whether to allow a point to be its own neighbor.
    show_progress : bool
        Whether to show a progress bar during computation.
    comm : mpi.Comm, optional
        The MPI communicator for parallel processing, by default None.
        
    Returns
    -------
    List[List[int]]
        A list of lists, where each inner list contains the indices of
        neighboring points for a given point.
        
    Examples
    --------
    >>> coords = np.array([[0.0, 0.0], [1.0, 0.0], [0.0, 1.0], [5.0, 5.0]])
    >>> neighbor_list = make_neighbor_list_naive(coords, radius=1.5, 
    ...                                         allow_selfloop=False, 
    ...                                         show_progress=False)
    >>> neighbor_list
    [[1, 2], [0, 2], [0, 1], []]
    """
    if comm is None:
        mpisize = 1
        mpirank = 0
    else:
        mpisize = comm.size
        mpirank = comm.rank

    npoints = X.shape[0]
    points = np.array_split(range(npoints), mpisize)[mpirank]
    npoints_local = len(points)
    nnlist: List[List[int]] = [[] for _ in range(npoints_local)]

    if mpirank == 0 and show_progress and has_tqdm:
        desc = "rank 0" if mpisize > 1 else None
        ns = tqdm(points, desc=desc)
    else:
        ns = points

    for n in ns:
        xs = X[n, :]
        for m in range(npoints):
            if (not allow_selfloop) and n == m:
                continue
            ys = X[m, :]
            r = np.linalg.norm(xs - ys)
            if r <= radius:
                nnlist[n - points[0]].append(m)
    if mpisize > 1:
        nnlist = list(itertools.chain.from_iterable(comm.allgather(nnlist)))

    nnlist = [sorted(nn) for nn in nnlist]
    return nnlist


def make_neighbor_list(
    X: np.ndarray,
    radius: float,
    allow_selfloop: bool = False,
    check_allpairs: bool = False,
    show_progress: bool = False,
    comm: mpi.Comm = None,
) -> List[List[int]]:
    """
    Create a neighbor list for given points.
    
    This function serves as a unified interface to create neighbor lists,
    choosing between cell-based or naive implementation based on the parameters.
    
    Parameters
    ----------
    X : np.ndarray
        The coordinates of the points (N x D array where N is the number of points
        and D is the dimensionality).
    radius : float
        The radius within which points are considered neighbors.
    allow_selfloop : bool, optional
        Whether to allow a point to be its own neighbor, by default False.
    check_allpairs : bool, optional
        Whether to use the naive all-pairs approach instead of the cell-based one,
        by default False.
    show_progress : bool, optional
        Whether to show a progress bar during computation, by default False.
    comm : mpi.Comm, optional
        The MPI communicator for parallel processing, by default None.
        
    Returns
    -------
    List[List[int]]
        A list of lists, where each inner list contains the indices of
        neighboring points for a given point.
        
    Examples
    --------
    >>> coords = np.array([[0.0, 0.0], [1.0, 0.0], [0.0, 1.0], [5.0, 5.0]])
    >>> neighbor_list = make_neighbor_list(coords, radius=1.5)
    >>> neighbor_list
    [[1, 2], [0, 2], [0, 1], []]
    
    >>> # Force all-pairs algorithm
    >>> neighbor_list = make_neighbor_list(coords, radius=1.5, check_allpairs=True)
    >>> neighbor_list
    [[1, 2], [0, 2], [0, 1], []]
    """
    if check_allpairs:
        return make_neighbor_list_naive(
            X,
            radius,
            allow_selfloop=allow_selfloop,
            show_progress=show_progress,
            comm=comm,
        )
    else:
        return make_neighbor_list_cell(
            X,
            radius,
            allow_selfloop=allow_selfloop,
            show_progress=show_progress,
            comm=comm,
        )


def load_neighbor_list(filename: PathLike, nnodes: int = None) -> List[List[int]]:
    """
    Load a neighbor list from a file.
    
    The file format is expected to have one line per node, with the first number
    being the node index and subsequent numbers being its neighbors.
    
    Parameters
    ----------
    filename : PathLike
        The path to the file containing the neighbor list.
    nnodes : int, optional
        The number of nodes. If None, it will be determined from the file
        by counting the number of non-empty lines.
        
    Returns
    -------
    List[List[int]]
        The neighbor list loaded from the file.
        
    Examples
    --------
    >>> # Example file content:
    >>> # 0 1 2
    >>> # 1 0 2
    >>> # 2 0 1
    >>> neighbor_list = load_neighbor_list("neighborlist.txt")
    >>> neighbor_list
    [[1, 2], [0, 2], [0, 1]]
    """
    if nnodes is None:
        nnodes = 0
        with open(filename) as f:
            for line in f:
                line = line.split("#")[0].strip()
                if len(line) == 0:
                    continue
                nnodes += 1

    neighbor_list: List[List[int]] = [[] for _ in range(nnodes)]
    with open(filename) as f:
        for line in f:
            line = line.strip().split("#")[0]
            if len(line) == 0:
                continue
            words = line.split()
            i = int(words[0])
            nn = [int(w) for w in words[1:]]
            neighbor_list[i] = nn
    return neighbor_list


def write_neighbor_list(
    filename: str,
    nnlist: List[List[int]],
    radius: float = None,
    unit: np.ndarray = None,
) -> None:
    """
    Write the neighbor list to a file.
    
    The file format has one line per node, with the first number being the 
    node index and subsequent numbers being its neighbors. Optional metadata
    can be included as comments at the beginning of the file.
    
    Parameters
    ----------
    filename : str
        The path to the output file.
    nnlist : List[List[int]]
        The neighbor list to write.
    radius : float, optional
        The neighborhood radius, written as a comment in the file, by default None.
    unit : np.ndarray, optional
        The unit lengths for each coordinate dimension, written as a comment
        in the file, by default None.
        
    Returns
    -------
    None
        This function writes the neighbor list to a file.
        
    Examples
    --------
    >>> neighbor_list = [[1, 2], [0, 2], [0, 1]]
    >>> write_neighbor_list("neighborlist.txt", neighbor_list, radius=1.5,
    ...                    unit=np.array([1.0, 1.0, 1.0]))
    """
    with open(filename, "w") as f:
        if radius is not None:
            f.write(f"# radius = {radius}\n")
        if unit is not None:
            f.write(f"# unit =")
            for u in unit:
                f.write(f" {u}")
            f.write("\n")
        for idx, nn in enumerate(nnlist):
            f.write(" ".join(map(str, [idx, *nn])) + "\n")

def main() -> None:
    """
    Command-line utility for creating neighbor lists from mesh data files.
    
    This function parses command-line arguments and creates a neighbor list file
    from a given input mesh file.
    
    Returns
    -------
    None
        This function is the main entry point for the command-line utility.
    """
    import argparse

    parser = argparse.ArgumentParser(
        description="Make neighbor-list file from mesh-data file",
        epilog="""
Note:
  - The first column of an input file will be ignored.
  - UNIT is used for changing aspect ratio (a kind of normalization)
    - Each coodinate will be divided by the corresponding unit
    - UNIT is given as a string separated with white space
      - For example, -u "1.0 0.5 1.0" for 3D case
  - tqdm python package is required to show progress bar
""",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument("input", type=str, help="input file")
    parser.add_argument(
        "-o", "--output", type=str, default="neighborlist.txt", help="output file"
    )
    parser.add_argument(
        "-r", "--radius", type=float, default=1.0, help="neighborhood radius"
    )
    parser.add_argument(
        "-u",
        "--unit",
        default=None,
        help="length unit for each coordinate (see Note)",
    )
    parser.add_argument(
        "-q", "--quiet", action="store_true", help="Do not show progress bar"
    )
    parser.add_argument(
        "--progress", action="store_true", help="show progress bar"
    )
    parser.add_argument("--allow-selfloop", action="store_true", help="allow self loop")
    parser.add_argument(
        "--check-allpairs",
        action="store_true",
        help="check all pairs (bruteforce algorithm)",
    )

    args = parser.parse_args()

    inputfile = args.input
    outputfile = args.output
    radius = args.radius

    if (args.progress or not args.quiet) and not has_tqdm:
        print("WARNING: cannot show progress because tqdm package is not available")

    X = np.zeros((0, 0))

    if mpi.rank() == 0:
        X = np.loadtxt(inputfile)

    if mpi.size() > 1:
        sh = mpi.comm().bcast(X.shape, root=0)
        if mpi.rank() != 0:
            X = np.zeros(sh)
        mpi.comm().Bcast(X, root=0)

    D = X.shape[1] - 1

    if args.unit is None:
        unit = np.ones(D, dtype=float)
    else:
        unit = np.array([float(w) for w in args.unit.split()])
        if len(unit) != D:
            print(f"--unit option expects {D} floats as a string but {len(unit)} given")
            sys.exit(1)
    X = X[:, 1:] / unit

    nnlist = make_neighbor_list(
        X,
        radius,
        allow_selfloop=args.allow_selfloop,
        check_allpairs=args.check_allpairs,
        show_progress=(args.progress or not args.quiet),
        comm=mpi.comm(),
    )

    write_neighbor_list(outputfile, nnlist, radius=radius, unit=unit)



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import numpy as np

import odatse
import odatse.solver.function

def quadratics(xs: np.ndarray) -> float:
    """
    Quadratic (sphere) function.

    Parameters
    ----------
    xs : np.ndarray
        Input array.

    Returns
    -------
    float
        The calculated value of the quadratic function.

    Notes
    -----
    It has one global minimum f(xs)=0 at xs = [0,0,...,0].
    """
    return np.sum(xs * xs)

def quartics(xs: np.ndarray) -> float:
    """
    Quartic function with two global minima.

    Parameters
    ----------
    xs : np.ndarray
        Input array.

    Returns
    -------
    float
        The calculated value of the quartic function.

    Notes
    -----
    It has two global minima f(xs)=0 at xs = [1,1,...,1] and [0,0,...,0].
    It has one saddle point f(0,0,...,0) = 1.0.
    """
    return np.mean((xs - 1.0) ** 2) * np.mean((xs + 1.0) ** 2)


def ackley(xs: np.ndarray) -> float:
    """
    Ackley's function in arbitrary dimension

    Parameters
    ----------
    xs : np.ndarray
        Input array.

    Returns
    -------
    float
        The calculated value of Ackley's function.

    Notes
    -----
    It has one global minimum f(xs)=0 at xs=[0,0,...,0].
    It has many local minima.
    """
    a = np.mean(xs ** 2)
    a = 20 * np.exp(-0.2 * np.sqrt(a))
    b = np.cos(2.0 * np.pi * xs)
    b = np.exp(0.5 * np.sum(b))
    return 20.0 + np.exp(1.0) - a - b

def rosenbrock(xs: np.ndarray) -> float:
    """
    Rosenbrock's function.

    Parameters
    ----------
    xs : np.ndarray
        Input array.

    Returns
    -------
    float
        The calculated value of Rosenbrock's function.

    Notes
    -----
    It has one global minimum f(xs) = 0 at xs=[1,1,...,1].
    """
    return np.sum(100.0 * (xs[1:] - xs[:-1] ** 2) ** 2 + (1.0 - xs[:-1]) ** 2)

def himmelblau(xs: np.ndarray) -> float:
    """
    Himmelblau's function.

    Parameters
    ----------
    xs : np.ndarray
        Input array of shape (2,).

    Returns
    -------
    float
        The calculated value of Himmelblau's function.

    Notes
    -----
    It has four global minima f(xs) = 0 at
    xs=[3,2], [-2.805118..., 3.131312...], [-3.779310..., -3.2831860], and [3.584428..., -1.848126...].
    """
    if xs.shape[0] != 2:
        raise RuntimeError(
            f"ERROR: himmelblau expects d=2 input, but receives d={xs.shape[0]} one"
        )
    return (xs[0] ** 2 + xs[1] - 11.0) ** 2 + (xs[0] + xs[1] ** 2 - 7.0) ** 2

def linear_regression_test(xs: np.ndarray) -> float:
    """
    Negative log likelihood of linear regression with Gaussian noise N(0,sigma)

    y = ax + b

    trained by xdata = [1, 2, 3, 4, 5, 6] and ydata = [1, 3, 2, 4, 3, 5].

    Model parameters (a, b, sigma) are corresponding to xs as the following,
    a = xs[0], b = xs[1], log(sigma**2) = xs[2]

    It has a global minimum f(xs) = 1.005071.. at
    xs = [0.628571..., 0.8, -0.664976...].

    Parameters
    ----------
    xs : np.ndarray
        Input array of model parameters.

    Returns
    -------
    float
        The negative log likelihood of the linear regression model.
    """
    if xs.shape[0] != 3:
        raise RuntimeError(
            f"ERROR: regression expects d=3 input, but receives d={xs.shape[0]} one"
        )

    xdata = np.array([1, 2, 3, 4, 5, 6])
    ydata = np.array([1, 3, 2, 4, 3, 5])
    n = len(ydata)

    return 0.5 * (
        n * xs[2] + np.sum((xs[0] * xdata + xs[1] - ydata) ** 2) / np.exp(xs[2])
    )

class Solver(odatse.solver.function.Solver):
    """Function Solver with pre-defined benchmark functions"""

    x: np.ndarray
    fx: float

    def __init__(self, info: odatse.Info) -> None:
        """
        Initialize the solver.

        Parameters
        ----------
        info: Info
            Information object containing solver configuration.
        """
        super().__init__(info)
        self._name = "analytical"
        function_name = info.solver.get("function_name", "quadratics")

        if function_name == "quadratics":
            self.set_function(quadratics)
        elif function_name == "quartics":
            self.set_function(quartics)
        elif function_name == "ackley":
            self.set_function(ackley)
        elif function_name == "rosenbrock":
            self.set_function(rosenbrock)
        elif function_name == "himmelblau":
            dimension = self.dimension
            if int(dimension) != 2:
                raise RuntimeError(
                    f"ERROR: himmelblau works only with dimension=2 but input is dimension={dimension}"
                )
            self.set_function(himmelblau)
        elif function_name == "linear_regression_test":
            dimension = self.dimension
            if int(dimension) != 3:
                raise RuntimeError(
                    f"ERROR: regression works only with dimension=2 but input is dimension={dimension}"
                )
            self.set_function(linear_regression_test)
        else:
            raise RuntimeError(f"ERROR: Unknown function, {function_name}")



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

from .meshgrid import MeshGrid
from .region import Region
